unit UnitFrmInstance;
{
    Purpose:
        Recieve data from a previous instance;

    Notes:
        This form is conditionally created for only the first instance
        of this program.        
}

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs;


const MYMESSAGE = WM_USER + 5;
const MYCLASS = 'ArsClip.FrmDummyInstance';
type InstanceCommand = (INST_INIT, INST_POPUP, INST_PARAM);
type
  TFrmDummyInstance = class(TForm)
  private
    mycommandline : string;
    params : array[0 .. 31] of string;
    pcount : integer;

    { Private declarations }

    command : InstanceCommand;

  public
    procedure CreateParams(var Params: TCreateParams); override;
    { Public declarations }
    procedure GetDataMessage(var Msg: TWMCopyData); message WM_COPYDATA;

  end;




var
  FrmDummyInstance: TFrmDummyInstance;

procedure FrmInstance_SendData(data : string);
procedure FrmInstance_Init(ic : InstanceCommand);  overload;
procedure FrmInstance_Init(p : string);  overload;
procedure FrmInstance_DeInit;
function FrmInstance_IsFirst : boolean;

implementation

uses  StrUtils, UnitFrmMainPopup, UnitPopupGenerate, UnitFrmJumpList;


{$R *.dfm}
var mutex : THandle;
    First : Boolean;
    param : string;
{ TFrmDummyInstance }

function FrmInstance_IsFirst : boolean;
begin
    result := First;
end;

procedure FrmInstance_Init(p : string);
begin
    param := p;
    FrmInstance_init(INST_PARAM);
end;

procedure FrmInstance_Init(ic : InstanceCommand);
begin
    mutex := Windows.CreateMutex(nil, true, pchar(ExtractFileName(Application.exename)));
    first := (Windows.GetLastError = 0);

    case ic of
        INST_INIT: if First then Application.CreateForm(TFrmDummyInstance, FrmDummyInstance);
        INST_POPUP:
            begin
                Application.Terminate;
                Application.ProcessMessages;
                UnitFrmInstance.FrmInstance_SendData(IntToStr(Integer(ic)));
                Application.ProcessMessages;
            end;
        INST_PARAM:
            begin
                Application.Terminate;
                Application.ProcessMessages;
                UnitFrmInstance.FrmInstance_SendData(param);
                Application.ProcessMessages;
            end;
    end;

end;

procedure FrmInstance_DeInit;
begin
    if mutex <> 0 then begin
        Windows.CloseHandle(mutex);
    end;
end;

procedure FrmInstance_SendData(data : string);
var
  cd : TCopyDataStruct;
  s, param : string;
  h : THandle;
  i : integer;
begin
    s := data + #0;

    try

        ParamCount;
        cd.dwData := 0;
        cd.cbData := length(s);
        cd.lpData := StrAlloc(cd.cbData);
        StrPCopy(PChar(cd.lpData), s);

        // Send copy data message
        // set result according to result returned from other instance
        h := FindWindow(MYCLASS, nil);


        if (h <> 0) then begin
            SendMessage(
                h,
                WM_COPYDATA,
                0,
                integer(@cd)
            );
        end;

    finally
    end;
end;


procedure TFrmDummyInstance.GetDataMessage(var Msg: TWMCopyData);
var s : string;
begin
    s := PChar(MSg.CopyDataStruct.lpData);
    command := INST_PARAM;
    try
        command := InstanceCommand(StrToInt(s));
    except
        // the PARAM isn't an integer and will cause an exception
    end;

    case command of
    INST_INIT: ;
    INST_POPUP:
        begin
            FrmMainPopup.FakePopupHotkey;
        end;
    INST_PARAM:
        begin
            FrmJumpList.HandleTaskArgument(s);
        end;
    end;
end;

procedure TFrmDummyInstance.CreateParams(var Params: TCreateParams);
begin
    inherited;

    StrCopy(Params.WinClassName, MYCLASS);
end;


end.
