unit UnitFrmPasteSelected;
{
    Purpose:
        Display the clipboard
        User select 0 or more items
        All items are returned as a string - empty if none selected

    Updates:

        Fixed cancel button
        ------------

        Added option to reverse pasting order
        Using UnitClipQueue
}
interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, UnitTWideChar;

type
  TFrmPasteSelected = class(TForm)
    lbHistory: TListBox;
    btnPaste: TButton;
    btnCancel: TButton;
    cbPasteReverse: TCheckBox;
    procedure FormShow(Sender: TObject);
    procedure btnPasteClick(Sender: TObject);
    procedure btnCancelClick(Sender: TObject);
    procedure FormKeyUp(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormActivate(Sender: TObject);
    procedure lbHistoryClick(Sender: TObject);
    procedure lbHistoryMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure lbHistoryMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure lbHistoryKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure lbHistoryDrawItem(Control: TWinControl; Index: Integer;
      Rect: TRect; State: TOwnerDrawState);
  private
    { Private declarations }
    Cancelled : boolean;
    mysel : array of boolean;
    shifted : boolean;
    shiftedIdx : integer;
    lastIdx : integer;
  public
    { Public declarations }
    function GetSelectedItems() : string; overload;
    function GetSelectedItemsWide() : TWideChar; overload;
  end;

var
  FrmPasteSelected: TFrmPasteSelected;

implementation

uses UnitFrmClipboardManager, UnitFrmMainPopup, UnitClipQueue , UnitMisc;

{$R *.dfm}

//----------------------------------
// public interface
//----------------------------------
function TFrmPasteSelected.GetSelectedItems() : string;
var i : longint;
    s : string;
begin
    //
    // We need to steal focus to use this windows after the popup
    // ;otherwise, it will pop-under on the second call

    Windows.SetForegroundWindow(self.handle);
    self.ShowModal;

    s := '';
    if (not self.Cancelled) then begin
        if cbPasteReverse.Checked then begin
            for i := (lbHistory.items.count - 1) downto 0 do begin
                if lbHistory.Selected[i] then begin
                    if s = '' then begin
                        s := s + lbHistory.items[i];
                    end else begin
                        s := s + #13#10 + lbHistory.items[i];
                    end;
                end;
            end;
        end else begin
            for i := 0 to (lbHistory.items.count - 1) do begin
                if lbHistory.Selected[i] then begin
                    if s = '' then begin
                        s := s + lbHistory.items[i];
                    end else begin
                        s := s + #13#10 + lbHistory.items[i];
                    end;
                end;
            end;
        end;
    end;
    result := s;
end;

function TFrmPasteSelected.GetSelectedItemsWide: TWideChar;
var i, start, stop : longint;
    s : string;
    ci : TClipItem;
begin
    //
    // We need to steal focus to use this windows after the popup
    // ;otherwise, it will pop-under on the second call

    Windows.SetForegroundWindow(self.handle);
    self.ShowModal;

    s := '';
    result := nil;
    if (not self.Cancelled) then begin
        result := TWideChar.Create;
        if cbPasteReverse.Checked then begin
            for i := (lbHistory.items.count - 1) downto 0 do begin
                if lbHistory.Selected[i] then begin
                    ci := ClipQueue.GetClipItem(i);
                    if result.size <> 0 then begin
                        result.Append(#13#10);
                    end;

                    if ci.GetFormat = CF_UNICODETEXT then begin
                        result.Append(ci.GetHandle, ci.GetDataSize);
                    end else begin
                        result.Append(ci.GetAsText);
                    end;
                end;
            end;
        end else begin
            for i := 0 to (lbHistory.items.count - 1) do begin
                if lbHistory.Selected[i] then begin
                    ci := ClipQueue.GetClipItem(i);
                    if result.size <> 0 then begin
                        result.Append(#13#10);
                    end;

                    if ci.GetFormat = CF_UNICODETEXT then begin
                        result.Append(ci.GetHandle, ci.GetDataSize);
                    end else begin
                        result.Append(ci.GetAsText);
                    end;
                end;
            end;
        end;
        if result.size = 0 then result := nil;
    end;

    
    //result := s;
end;

//----------------------------------
// private implemenation
//----------------------------------

procedure TFrmPasteSelected.FormShow(Sender: TObject);
begin
    // get the latest list
    lbHistory.Items.Clear;
    ClipQueue.GetQueueItems(lbHistory.items);

end;

procedure TFrmPasteSelected.btnPasteClick(Sender: TObject);
begin
    self.Cancelled := false;
    self.close;
end;

procedure TFrmPasteSelected.btnCancelClick(Sender: TObject);
begin
    self.Cancelled := true;
    self.close;
end;




procedure TFrmPasteSelected.FormKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
    if (KEY = VK_ESCAPE) then begin
        self.btnCancel.Click;
    end;
end;

procedure TFrmPasteSelected.FormActivate(Sender: TObject);
begin
    self.lbHistory.SetFocus;
    setlength(mysel, lbHistory.Count);
end;

procedure TFrmPasteSelected.lbHistoryClick(Sender: TObject);
var i, idx : integer;
begin
    with lbhistory do begin
        idx := ItemIndex;
        if shifted then begin
            for i := 0 to (Count - 1) do begin
                if (lastidx < shiftedIdx) then begin
                    mysel[i] := (i >= lastidx) and (i <= shiftedIdx);
                end else begin
                    mysel[i] := (i >= shiftedIdx) and (i <= lastidx);
                end;
            end;
        end else begin
            mysel[ItemIndex] := not mysel[ItemIndex];
        end;

         for i := 0 to lbhistory.Count - 1 do begin
            lbhistory.Selected[i] := mysel[i];
         end;
         ItemIndex := idx;
    end;
end;

procedure TFrmPasteSelected.lbHistoryDrawItem(Control: TWinControl;
  Index: Integer; Rect: TRect; State: TOwnerDrawState);
var wc : TWideChar;
    ci : TClipItem;
    c : TColor;
begin

    c := tlistbox(control).Canvas.Pen.Color;
    if tlistbox(control).Selected[index]  then begin
        tlistbox(control).Canvas.Pen.Color := clHighlight;
        tlistbox(control).canvas.Rectangle(rect);
    end;
    tlistbox(control).Canvas.Pen.Color := c;
    
    if UnitMisc.UniSupported then begin

        ci := ClipQueue.GetClipItem(index);
        wc := TWideChar.Create;
        if ci.GetFormat = CF_UNICODETEXT then begin
            wc.Append(ci.GetHandle, ci.GetDataSize);
        end else begin
            wc.Append(ci.GetAsText);
        end;
        Windows.TextOutW(
            TListBox(Control).Canvas.Handle,
            rect.Left, rect.Top, wc.GetMemoryStream.Memory, wc.StrLength
            );
    end else begin
        Windows.TextOut(
            TListBox(Control).Canvas.Handle,
            rect.Left, rect.Top, pchar(lbHistory.items[index]) , length(lbHistory.Items[index])
            );
    end;
end;

procedure TFrmPasteSelected.lbHistoryMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
    lbHistory.Items.BeginUpdate;
    shifted := ssShift in shift;
    if shifted then begin
        shiftedIdx := lbHistory.ItemIndex;
    end else begin
        lastIdx := lbHistory.ItemIndex;
    end;
end;

procedure TFrmPasteSelected.lbHistoryMouseUp(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
    lbHistory.Items.EndUpdate;
end;

procedure TFrmPasteSelected.lbHistoryKeyDown(Sender: TObject;
  var Key: Word; Shift: TShiftState);
//var idx, i : integer;
begin
    {
    with lbhistory do begin
        if key = VK_SPACE then begin
        idx := ItemIndex;
            mysel[ItemIndex] := not mysel[ItemIndex];
         for i := 0 to lbhistory.Count - 1 do begin
            lbhistory.Selected[i] := mysel[i];
         end;
         ItemIndex := idx;
        end;
    end;
    }
end;

end.
