unit UnitFrmPermanentNew;
{
    Purpose:
        This unit stores/reads/edits the permanent items.
        The form is not a dummy form.


    Updates:
        New: Permanent Items Groups sorted by name
        --------------
        Fix: Up button did not range check
        --------------
        Changes for complex item as permanent items,
        show them when it's shown or hovered

        Folder list not cleared when "LoadPermanent" was called
        ----------------
        Made scalable
        New Keystroke support

        -----------------
        Updates for autotamically switching PIGs for a specific program

        -------------------
        Complete Rewrite. Old form was too poorly designed
        to save.


}

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls, INIFiles, ComCtrls {for THahsedStringList},
  UnitClipQueue, UnitFrmDummyUnicodeTooltip, Menus, Buttons, ClipBrd, Contnrs;

const DEFAULT_FOLDER = 'Default';
      ADDNEW_FOLDER = '<add new>';
      PERM0_FILE = 'perm0.ini';
      PERM1_FILE = 'perm1.ini';
      HOTKEY_FILE = 'hotkey.ini';
      EXEPIG_FILE = 'exepigs.ini';

type TDataFileName = (DFN_PERM0, DFN_PERM1, DFN_HOTKEYS);

type THotkeyData = record
    alt : boolean;
    ctrl : boolean;
    shft : boolean;
    win : boolean;
    key : word;
    name : string;
end;


type
  TFrmPermanent = class(TForm)
    pcPermanent: TPageControl;
    TabSheet2: TTabSheet;
    lvAutoSwitch: TListView;
    Label5: TLabel;
    pmGetAs: TPopupMenu;
    PlaintText1: TMenuItem;
    DIBPicture1: TMenuItem;
    CopiedFiles1: TMenuItem;
    RichTExt1: TMenuItem;
    tsEditGroups: TTabSheet;
    tvGroups: TTreeView;
    btnNewItem: TButton;
    btnEditItem: TButton;
    TabSheet3: TTabSheet;
    Label7: TLabel;
    Label8: TLabel;
    Label9: TLabel;
    pnlText: TPanel;
    imgPic: TImage;
    memText: TMemo;
    reText: TRichEdit;
    pnlHotkey: TPanel;
    lblHotkey: TLabel;
    Label11: TLabel;
    Shape2: TShape;
    pmTree: TPopupMenu;
    miDelete: TMenuItem;
    miMoveto: TMenuItem;
    procedure FormCreate(Sender: TObject);
    procedure FormShow(Sender: TObject);
//    procedure cbPIGsChange(Sender: TObject);
//    procedure btnDeleteClick(Sender: TObject);
//    procedure btnDownClick(Sender: TObject);
//    procedure btnUpClick(Sender: TObject);
//    procedure btnEditClick(Sender: TObject);
//    procedure lbItemNamesClick(Sender: TObject);
//    procedure btnCancelClick(Sender: TObject);
//    procedure btnSaveClick(Sender: TObject);
    procedure btnAddPIGClick(Sender: TObject);
//    procedure btnNewClick(Sender: TObject);
//    procedure btnDeletePIGClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
//    procedure cbKeystrokesClick(Sender: TObject);
//    procedure Button1Click(Sender: TObject);
//    procedure bGetClipboardClick(Sender: TObject);
//    procedure bGetClipboardAsClick(Sender: TObject);
//    procedure btnMoveClick(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
//    procedure bEditHotkeyClick(Sender: TObject);
//    procedure btnClearClick(Sender: TObject);
    procedure tsEditGroupsShow(Sender: TObject);
    procedure tvGroupsExpanding(Sender: TObject; Node: TTreeNode;
      var AllowExpansion: Boolean);
    procedure tvGroupsMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure tvGroupsDragOver(Sender, Source: TObject; X, Y: Integer;
      State: TDragState; var Accept: Boolean);
    procedure tvGroupsEndDrag(Sender, Target: TObject; X, Y: Integer);
    procedure tvGroupsMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure tvGroupsClick(Sender: TObject);
    procedure tvGroupsCollapsing(Sender: TObject; Node: TTreeNode;
      var AllowCollapse: Boolean);
    procedure btnNewItemClick(Sender: TObject);
//    procedure btnEditItemClick(Sender: TObject);
    procedure miDeleteClick(Sender: TObject);
    procedure btnEditItemClick(Sender: TObject);
    procedure miMovetoItemClick(Sender: TObject);

    //procedure Button1Click(Sender: TObject);
  private
  { Private declarations }
        SelectedIndex : integer;
        SelectedGroup : string;
        AppPath : string;
        CurrentPermPath : string;
        FolderList : TStringList;
        ItemNameList : TStringList;
        ItemDataList : TStringList;
        ItemHotKeyList : TStringList;
        EXEPigList : THashedStringList;
        EXEPasteList : THashedStringList;
        Tooltip : TTooltipWindow;
        Clipboard : TClipItem;

        HotkeyObjects : TStringList;
        HotkeyObjects2 : TStringList;
        PushpopPath : string;
        EditModeOn : boolean;
        hkd : THotkeyData;

        //treenodedrag : TTreeNode;
        accepteddrag : boolean;
        // util functions
        function GetOldDataFilename (i: integer): string;
        function GetDataFilename (i: TDataFileName): string;

        function ToString(hkd : THotkeyData) : string;
        function FromString(s : string) : THotKeyData;
        procedure UpdateFolderList;
        procedure LoadPermanentItemsGroup;
        procedure SavePermanentItemsGroup;
        function IsPermanentPathFolder(folder : string) : boolean;
        procedure RefreshFormData;
        //
//        procedure SetEditMode(value : boolean);
//        procedure ShowSelectedItem;
//        procedure RefreshCurrentItem;

//        procedure PopupItemClick(sender: TObject);

        function GetNextComplexIndex : integer;
        function ExtractComplexIndex(s : string) : integer;

        procedure RefreshHotkeys;
        procedure RefreshPermanentGroups(selectCurrentGroup : boolean);
        procedure SelectTreeItem(group : string; index : integer);
        procedure ShowSelectedTreeItem;
  public
  { Public declarations }
        procedure SetPermanentPath( path : string );
        function GetPermanentPath : string;
        function GetPermanentPathFull : string;
        // to enumerate permanent item names
        function PermFoldersGetCount : cardinal;
        function PermFoldersGetItem(index : cardinal) : string;
        function IsComplexItem(s : string) : boolean;
        function GetComplexItem(s : string) : TClipItem;

        procedure SaveNewItem(
            group : string;
            name, text : string;
            hotkey : THotKeyData;
            usekeys : boolean;
            ci : TClipItem = nil
            );
        procedure SaveItem(
            group : string; index : integer;
            name, text : string;
            hotkey : THotKeyData;
            usekeys : boolean;
            ci : TClipItem = nil
            );

        procedure PermFolderPush;
        procedure PermFolderPop;

        // to enumerate items in the current permanent folder
        function GetCount: integer;
        function GetItemName(i: integer): string;
        function GetItemText(i: integer): string;
        function GetItemHotkeyName(i : integer) : string;
        function GetTextFrom(name: string): string;
        function GetTextIndex(name : string) : integer;

        // display form with a newly created item
        procedure ShowWithNewItem(item : string; name : string = ''; hotkey : string = ''); overload;
        procedure ShowWithNewComplexItem(ci : TClipItem; name : string = ''; hotkey : string = ''); overload;
        procedure ShowAndSelect(groupID, itemID : integer);


        procedure AutoSwitch(EXEName : string);
        procedure AssignPIG(EXEName : string);

        function GetHotkeyCount : integer;
        function GetHotkeyIdent(index : integer) : integer;
        function IsHotkeyIdent(ident : integer) : boolean;
        procedure ReportHotkey(index : integer);
  end;

var
  FrmPermanent: TFrmPermanent;
const EMPTY_HOTKEY_NAME = 'none';
{////////////////////}
{//}implementation{//}
{////////////////////}

uses UnitMisc, StrUtils, UnitFrmClipboardManager, UnitFrmMove, UnitHotKey,
  UnitFrmHotKey, UnitFrmMainPopup, UnitPopupGenerate, UnitTWideChar,
  UnitFrmPermanentEdit;

{$R *.dfm}



//
// public API
//
procedure TfrmPermanent.SetPermanentPath( path : string );
begin
    if (self.EditModeOn) then begin
        ShowMessage('Cannot switch groups while editing a permanent item.');
        EXIT;
    end;

    if Self.IsPermanentPathFolder(path) then  begin
        CurrentPermPath := path;
        self.LoadPermanentItemsGroup;
    end else begin
        CurrentPermPath := DEFAULT_FOLDER;
        self.LoadPermanentItemsGroup;
    end;

    //lbItemNames.ItemIndex := -1;
end;

function TfrmPermanent.GetPermanentPath : string;
begin
    result := self.CurrentPermPath;
end;

function TFrmPermanent.GetPermanentPathFull: string;
begin
    result := IncludeTrailingPathDelimiter(
        IncludeTrailingPathDelimiter(self.AppPath) + self.GetPermanentPath
    );
end;


function TfrmPermanent.PermFoldersGetCount : cardinal;
begin
    result := FolderList.Count;
end;
function TfrmPermanent.PermFoldersGetItem(index : cardinal) : string;
begin
    result := FolderList[index];
end;


function TfrmPermanent.GetCount: integer;
begin
    result := ItemNameList.Count;
end;
function TfrmPermanent.GetItemName(i: integer): string;
begin
    result := ItemNameList[i];
end;
function TfrmPermanent.GetItemText(i: integer): string;
begin
    result := ItemDataList[i];
end;

function TFrmPermanent.GetItemHotkeyName(i: integer): string;
var hkd : THotkeyData;
begin
    hkd := FromString(ItemHotKeyList[i]);
    if hkd.name = EMPTY_HOTKEY_NAME then
        result := ''
    else
        result := hkd.name;
end;




function TFrmPermanent.GetTextIndex(name: string): integer;
var i: integer;
begin
    result := -1;
    for i := 0 to ItemNameList.count - 1 do begin
        if (lowercase(name) = lowercase(ItemNameList[i])) then begin
            result := i;
            BREAK;
        end;
    end;
end;

function TfrmPermanent.GetTextFrom(name: string): string;
begin
    result := ItemDataList[ GetTextIndex(name) ];
end;


procedure TFrmPermanent.ShowAndSelect(groupID, itemID: integer);
var s : string;
begin
    self.Show;

    Windows.SetForegroundWindow(self.Handle); 

    self.pcPermanent.ActivePageIndex := 0;

    s := self.CurrentPermPath;
    if groupID <> -1 then
        s := self.PermFoldersGetItem(groupID);
    self.SelectTreeItem(s, itemID);

end;

procedure TfrmPermanent.ShowWithNewItem(item : string; name : string = '';hotkey : string = '');
begin



    self.Show;
    self.pcPermanent.ActivePageIndex := 0;
    self.SelectTreeItem(self.CurrentPermPath, -1);
    FrmPermanentEdit.SetNewItem(item);
    self.btnNewItem.Click;

{
    mItemText.text := item;
    txtItemName.text := name;
    self.hkd := self.FromString(hotkey);
    self.SetEditMode(true);
    self.RefreshCurrentItem;
}
end;

procedure TFrmPermanent.ShowWithNewComplexItem(ci: TClipItem; name : string = '';hotkey : string = '');
var i : integer;
begin
    i := self.GetNextComplexIndex;
    if (ci <> nil) then begin
        ci.SaveToFile(self.GetPermanentPathFull, i );
    end;

    self.ShowWithNewItem('[FILE=' + IntToStr(i) + ']', name, hotkey);



{    self.Show;
    self.pcPermanent.ActivePageIndex := 0;
    self.SelectTreeItem(self.CurrentPermPath, -1);
    FrmPermanentEdit.SetNewItem(ci);
    self.btnNewItem.Click;
}
end;


//
// Create / Destroy
//

procedure TFrmPermanent.FormCreate(Sender: TObject);
var name: string;
    lc : TListColumn;

    procedure EmptyFile(name : string);
    var t : textfile;
    begin
        assignfile(t,name);
        rewrite(t);
        closefile(t);
    end;
begin
    UnitMisc.AppendLog('FrmPermanent - creating');

    self.Font.Size := 8;

    self.FolderList := TStringList.Create;
    self.ItemNameList := TStringList.Create;
    self.ItemDataList := TStringList.Create;
    self.ItemHotKeyList := TStringList.Create;
    self.EXEPigList := THashedStringList.Create;
    self.EXEPasteList := THashedStringList.Create;
    self.Tooltip := TTooltipWindow.Create;
    self.Clipboard := TClipItem.Create;
    self.HotkeyObjects := TStringList.Create;
    self.HotkeyObjects2 := TStringList.Create;

    self.CurrentPermPath := DEFAULT_FOLDER;
    self.AppPath := UnitMisc.GetAppPath;

    // fix
    if FileExists(self.AppPath + EXEPIG_FILE) then begin
        self.EXEPigList.LoadFromFile(self.AppPath + EXEPIG_FILE);
    end;

    //
    // make the new Default directory and import and
    // current permanent items
    //
    if not DirectoryExists( self.AppPath + DEFAULT_FOLDER) then begin
        mkdir(self.AppPath + DEFAULT_FOLDER);
        self.PermFolderPush;
        self.CurrentPermPath := DEFAULT_FOLDER;

        name := GetDataFilename(DFN_PERM0);
        EmptyFile(name);
        name := GetDataFilename(DFN_HOTKEYS);
        EmptyFile(name);
        name := GetDataFilename(dfn_perm1);
        EmptyFile(name);
        self.PermFolderPop;


        name := GetOldDataFilename(0);
        if FileExists(name) then
            copyfile(pchar(name), PChar(GetDataFileName(DFN_PERM0)), true);

        name := GetOldDataFilename(1);
        if fileExists(name) then
            copyfile(pchar(name), PChar(GetDataFilename(DFN_PERM1)), true);
    end;

    self.UpdateFolderList;
    self.RefreshHotkeys;

    //
    // init the Auto Switch window
    //
    lc := lvAutoSwitch.Columns.Add;
    lc.Caption := 'Program Name';
    lc.Width := 100;

    lc := lvAutoSwitch.Columns.Add;
    lc.Caption := 'Group';
    lc.Width := 100;

    //lblClipType.Caption := '';

    pcPermanent.ActivePageIndex := 0;
    pcPermanent.Align := alClient;

    //self.SetEditMode(false);
    self.PushpopPath := '';
end;

procedure TFrmPermanent.FormDestroy(Sender: TObject);
begin
    self.EXEPigList.SaveToFile(self.AppPath + EXEPIG_FILE );
    MyFree(self.EXEPasteList);
    MyFree(self.ItemNameList);
    MyFree(self.ItemDataList);
    MyFree(self.ItemHotkeyList);
    MyFree(self.FolderList);
    MyFree(self.Tooltip);
    MyFree(self.Clipboard);
end;
//
// Util Functions
//

function TfrmPermanent.GetOldDataFilename(i: integer): string;
begin
    result := self.AppPath + 'perm' + IntToStr(i) + '.ini';
end;
function TfrmPermanent.GetDataFilename(i: TDataFileName): string;
begin
    case i of
    DFN_PERM0 : result := IncludeTrailingPathDelimiter(self.AppPath + CurrentPermPath) + PERM0_FILE;
    DFN_PERM1 : result := IncludeTrailingPathDelimiter(self.AppPath + CurrentPermPath) + PERM1_FILE;
    DFN_HOTKEYS : result := IncludeTrailingPathDelimiter(self.AppPath + CurrentPermPath) + HOTKEY_FILE;
    end
end;




procedure TfrmPermanent.LoadPermanentItemsGroup;
var name, itemText, s : string;
    lineCount : cardinal;
    i : integer;
    tf : textfile;
begin

    // use default when none given
    // ensure the default exists
    if CurrentPermPath = '' then begin
        CurrentPermPath := DEFAULT_FOLDER;
    end;
    if not FileExists(self.AppPath + CurrentPermPath) then begin
        if CurrentPermPath = DEFAULT_FOLDER then
            ForceDirectories(self.AppPath + CurrentPermPath);
    end;



    //
    // load permanent items
    //
    ItemNameList.Clear;
    name := GetDataFilename(DFN_PERM0);
    if FileExists(name) then begin
        ItemNameList.LoadFromFile(name);
    end;

    ItemHotKeyList.Clear;
    name := GetDataFilename(DFN_HOTKEYS);
    if FileExists(name) then begin
        ItemHotKeyList.LoadFromFile(name);
    end;
    while ItemHotKeyList.Count < ItemNameList.Count do begin
        ItemHotKeyList.Add('');
    end;

    //
    // abort reading and show message on error
    // always close the file
    //
    ItemDataList.Clear;
    name := GetDataFilename(DFN_PERM1);
    if FileExists(name) then begin
        AssignFile(tf, name);
        Reset(tf, name);

        try
            while not eof(tf) do begin
                try
                    Readln(tf, s);
                    itemText := '';
                    lineCount := StrToInt(s);

                    for i := 0 to lineCount - 1 do begin
                        Readln(tf, s);
                        if (itemText = '') then begin
                            itemText := s;
                        end else begin
                            itemText := itemText + chr(13) + chr(10) + s;
                        end;
                    end;

                    ItemDataList.Add(itemText);
                except
                     on E: Exception do begin
                        ShowMessage('The "Permanent Item" file for group ' + CurrentPermPath + ' is corrupted - ' + name + #13#10#13#10 +
                                    'Error Message: ' + E.Message);
                        break;
                     end;
                end;
            end;
        finally
            CloseFile(tf);
        end;
    end;
end;




procedure TfrmPermanent.SavePermanentItemsGroup;
var name: string;
    s : string;
    cnt : cardinal;

    i,j: longint;
    tf: textfile;
begin
    if (CurrentPermPath = '') then
        EXIT;

    if not DirectoryExists(IncludeTrailingPathDelimiter(AppPath) + CurrentPermPath) then begin
        mkdir(IncludeTrailingPathDelimiter(AppPath) + CurrentPermPath);
    end;
    //
    // save items
    //
    name := GetDataFilename(DFN_PERM0);
    ItemNameList.SaveToFile(name);

    name := GetDataFilename(DFN_HOTKEYS);
    ItemHotKeyList.SaveToFile(name);

    name := GetDataFilename(dfn_perm1);
    AssignFile(tf, name);
    Rewrite(tf);


    for i := 0 to ItemDataList.Count - 1 do begin
        s := ItemDataList[i];

        cnt := 1;
        for j := 1 to length(s) - 1 do begin
            if (s[j] = #13) and (s[j+1]= #10) then inc(cnt);
        end;

        writeln(tf, cnt);
        writeln(tf, s);
    end;

    CloseFile(tf);
end;


procedure TfrmPermanent.UpdateFolderList;
var rec : TSearchRec;
    r : integer;
begin
    //
    // scan each subfolder and look for permanent item config files
    // this will generate a list of permanent item groups (using their
    // folder name)
    //
    FolderList.clear;
    FolderList.Sorted := true;
    r := FindFirst(AppPath + '*.*', faDirectory, rec);
    while (r = 0) do begin
        // is file a subfolder?
        if (rec.Attr and faDirectory) > 0 then begin
            if (rec.name <> '.') and (rec.name <> '..') then begin
                if IsPermanentPathFolder(rec.name) then begin
                    FolderList.Add(rec.name);
                end;
            end;
        end;

        r := FindNext(rec);
    end;
end;

function TFrmPermanent.IsPermanentPathFolder(folder : string) : boolean;
begin
    result := fileexists(IncludeTrailingPathDelimiter(AppPath) + folder + '\' + PERM0_FILE )
end;


procedure TFrmPermanent.FormShow(Sender: TObject);
begin
    self.LoadPermanentItemsGroup;
    self.RefreshFormData;
    //self.SetEditMode(false);
    //txtItemName.Text := '';
    //mItemText.Text := '';
    //lblKeyname.Caption := '';

    //self.btnEdit.Enabled := false;
    //self.btnDelete.Enabled := false;
    //self.btnUp.Enabled := false;
    //self.btnDown.Enabled := false;
    //self.btnMove.Enabled := false;
end;

procedure TFrmPermanent.RefreshFormData;
var i : integer;
    li : TListItem;
begin
    // show the curent Permanent Item Group names and
    // select the current folder in the dropdown
    {cbpigs.Clear;
    cbPIGs.Items.AddStrings(FolderList);
    for i := 0 to FolderList.Count - 1 do begin
        if lowercase(cbpigs.Items[i]) = lowercase(CurrentPermPath) then begin
            cbPIGs.ItemIndex := i;
        end;
    end;
    }
    // load the current item names
    //
    {lbItemNames.Clear;
    lbItemNames.Items.AddStrings(ItemNameList);
    }
    lvAutoSwitch.Items.clear;
    for i := 0 to EXEPigList.Count - 1 do begin
        li := lvAutoSwitch.items.Add;
        li.Caption := EXEPigList.Names[i];
        li.SubItems.Add(EXEPigList.Values[EXEPigList.Names[i]]);
    end;


    //
    // disable buttons
    //

   { btnUp.Enabled :=  lbItemNames.ItemIndex > -1;;
    btnDown.Enabled := btnUp.Enabled;
    btnEdit.Enabled := btnUp.Enabled;
    btnDelete.Enabled := btnUp.Enabled;
    btnMove.Enabled := btnUp.Enabled;}
end;



procedure TFrmPermanent.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
{    self.btnEdit.Enabled := false;
    self.btnDelete.Enabled := false;
    self.btnUp.Enabled := false;
    self.btnDown.Enabled := false;
    self.btnMove.Enabled := false;}
    Tooltip.CloseTooltip;
end;



procedure TFrmPermanent.btnEditItemClick(Sender: TObject);
var s : string;
    b : boolean;
    i : integer;
    tn : TTreeNode;
begin
    FrmPermanentEdit.top := self.top;
    FrmPermanentEdit.left := self.left;

    self.PermFolderPush;
    self.SetPermanentPath(self.SelectedGroup);
    s := ItemHotKeyList[self.selectedIndex];
    hkd := self.FromString(s);
    hkd.name := FrmHotKey.GetHumanReadable(
        hkd.alt,hkd.shft,hkd.ctrl,hkd.win,hkd.key, b
    );
    self.PermFolderPop;

    tn := tvGroups.Selected;
    FrmPermanentEdit.showform(
        false, self.SelectedGroup, self.SelectedIndex, hkd, nil
    );
    if FrmPermanentEdit.ModalResult <> mrCancel then begin
        self.CurrentPermPath := self.SelectedGroup;
        //tvgroups.Selected := tn;

        self.tsEditGroupsShow(self);
        //tvgroups.Selected := tn;
        self.ShowSelectedTreeItem;
    end;
end;

procedure TFrmPermanent.SelectTreeItem(group: string; index: integer);
var i : integer;
    tn : TTreeNode;
begin

    for i  := 0 to tvgroups.Items.count - 1 do begin
        tn := tvgroups.Items[i];
        if tn.Parent = nil then begin
            if tn.Text = group then begin
                tvgroups.Selected := tn;
                tn.Expand(false);
                if index >= 0 then tvgroups.Selected := tn.item[index];

                self.ShowSelectedTreeItem;

                BREAK;
            end;
        end;
    end;
end;



//
// add / delete Permanent Item Group
//
procedure TFrmPermanent.btnAddPIGClick(Sender: TObject);
var newgroup : string;
begin
    if
    Dialogs.InputQuery('Add Group','Please enter a new group name', newgroup)
    then begin
        if not IsPermanentPathFolder(newgroup) then begin
            self.CurrentPermPath  := newgroup;
            self.ItemNameList.Clear;
            self.ItemDataList.clear;
            self.SavePermanentItemsGroup;
            self.UpdateFolderList;
            self.RefreshFormData;
        end else begin
            Dialogs.showmessage('Group name already exists');
        end;
    end;
end;





procedure TFrmPermanent.AutoSwitch(EXEName: string);
var path : string;
begin
    // given an EXE name, change to a PIG if it's associated with one
    // Associate the current PIG with the EXE if no association exist

    path := EXEPigList.Values[EXEName];
    if (path <> '') then begin
        self.SetPermanentPath(path);
    end else begin
        if EXEName <> '' then begin
            EXEPigList.Values[EXEName] := self.GetPermanentPath;
        end;
    end;
end;



procedure TFrmPermanent.AssignPIG(EXEName: string);
begin
    EXEPigList.Values[EXEName] := self.GetPermanentPath;
end;





function TFrmPermanent.GetComplexItem(s: string): TClipItem;
var i : integer;
begin
    result := TClipItem.Create;

    s := midstr(s,7,length(s)-7);
    i := StrToInt(s);
    result.LoadFromFIle(
        self.GetPermanentPathFull,
        i);
end;

function TFrmPermanent.IsComplexItem(s: string): boolean;
begin
    result := uppercase(leftstr(s,6)) =  '[FILE=';
end;





procedure TFrmPermanent.PermFolderPop;
begin
    if self.PushpopPath <> self.CurrentPermPath then begin
        self.SetPermanentPath(self.PushpopPath);
    end;
    self.PushpopPath := '';
end;

procedure TFrmPermanent.PermFolderPush;
begin
    if self.PushpopPath <> '' then begin
        raise Exception.Create('ERROR: Permanent Item folder restore failed');
    end;
    self.PushpopPath := self.CurrentPermPath;
end;


function TFrmPermanent.GetNextComplexIndex: integer;
var i, j : integer;
    sl : TStringList;
begin
    sl := TStringList.Create;

    for i := 0 to (ItemDataList.Count - 1) do begin
        if self.IsComplexItem(ItemDataList[i]) then begin
            j := self.ExtractComplexIndex(ItemDataList[i]);
            sl.Add(IntToStr(j));
        end;
    end;

    result := 0;
    while sl.IndexOf(IntToStr(result)) <> -1 do inc(result);

    UnitMisc.MyFree(sl);
end;

function TFrmPermanent.ExtractComplexIndex(s: string): integer;
begin
    s := midstr(s,7,length(s)-7);
    result := StrToInt(s);
end;



procedure TFrmPermanent.FormCloseQuery(Sender: TObject;
  var CanClose: Boolean);
begin
{    if btnCancel.Visible then begin
        btnCancel.Click;
    end;
}
    CanClose := true;
end;




//
// Hotkey Management / Enumeration
//
function TFrmPermanent.GetHotkeyCount: integer;
begin
    result := HotkeyObjects.Count;
end;

function TFrmPermanent.GetHotkeyIdent(index: integer): integer;
begin
    result := TACHotKey(hotkeyobjects.Objects[index]).Getid;
end;

procedure TFrmPermanent.RefreshHotkeys;
var i, j, k : integer;
    s : string;
    hk : TACHotKey;
    hkd : THotkeyData;
begin
    for i := 0 to HotkeyObjects2.count - 1 do begin
        hk := TACHotKey(
            HotkeyObjects2.Objects[i]
        );
        //hk.UnSet;
        MyFree(hk);
    end;
    HotkeyObjects2.Clear;
    HotkeyObjects.Clear;
    self.PermFolderPush;
    k := 0;
    for i := 0 to self.PermFoldersGetCount - 1 do begin
        self.SetPermanentPath(self.PermFoldersGetItem(i));

        for j := 0 to self.GetCount - 1 do begin
            s := ItemHotKeyList[j];
            if s <> '' then begin
                hkd := self.FromString(s);
                try
                    if (hkd.name <> EMPTY_HOTKEY_NAME) and
                        (hkd.alt or hkd.ctrl or hkd.shft or hkd.win) then begin

                        hk := TACHotKey.Create(
                            frmmainpopup.Handle,
                            'ArsClip.PermItem' + hkd.name,
                            hkd.ctrl, hkd.alt, hkd.shft, hkd.win, hkd.key
                        );
                        HotkeyObjects.Values[IntToStr(hk.GetID)] := IntToHex(i,8) + IntToHex(j,8);

                        HotkeyObjects2.AddObject(IntToStr(k), hk);
                        inc(k);
                    end;
                except
                end;
            end;
        end;
    end;

    self.PermFolderPop;
end;


function TFrmPermanent.ToString(hkd: THotkeyData): string;
begin
    result :=
        IntToStr(Integer(hkd.alt)) +
        IntToStr(Integer(hkd.ctrl)) +
        IntToStr(Integer(hkd.shft)) +
        IntToStr(Integer(hkd.win)) +
        IntToHex(hkd.key,2) +
        hkd.name;

end;



function TFrmPermanent.FromString(s: string): THotKeyData;
begin
    result.name := EMPTY_HOTKEY_NAME;
    if length(s) < 6 then begin
        EXIT;
    end;

    result.alt := Boolean(StrToInt(s[1]));
    result.ctrl  := Boolean(StrToInt(s[2]));
    result.shft  := Boolean(StrToInt(s[3]));
    result.win  := Boolean(StrToInt(s[4]));
    result.key := StrToInt('$' +  copy(s,5,2));
    delete(s,1,6);
    result.name := s;
end;

function TFrmPermanent.IsHotkeyIdent(ident: integer): boolean;
begin
    result := HotkeyObjects.Values[IntToStr(ident)] <> '';
end;

procedure TFrmPermanent.ReportHotkey(index: integer);
var s : string;
    i,j : integer;
begin
    s := HotkeyObjects.Values[IntToStr(index)];
    if (s <> '') then begin
        i := StrToInt('$' + copy(s,1,8));
        j := StrToInt('$' + copy(s,9,8));

        self.PermFolderPush;
        self.SetPermanentPath( self.PermFoldersGetItem(i) );
        s := self.GetItemText(j);
        FrmMainPopup.SendText(s);
        self.PermFolderPop;
    end;
end;




//===========================================================
//  new Edit form
//===========================================================

const NEW_GROUP_TEXT = '[New Group ...]';

procedure TFrmPermanent.tsEditGroupsShow(Sender: TObject);
begin
    self.RefreshPermanentGroups(true);
end;




procedure TFrmPermanent.tvGroupsClick(Sender: TObject);
begin
    if tvgroups.Selected = nil then EXIT;
    if tvgroups.Selected.Text = NEW_GROUP_TEXT  then begin
        self.btnAddPIGClick(self);
        self.tsEditGroupsShow(self);
    end;

    btnNewItem.Enabled := false;
    btnEditItem.Enabled := false;
    if tvgroups.selected.Parent = nil then begin
        btnNewItem.Enabled := true;
    end else begin
        btnEditItem.Enabled := true
    end;
end;

procedure TFrmPermanent.tvGroupsCollapsing(Sender: TObject; Node: TTreeNode;
  var AllowCollapse: Boolean);
begin
   { AllowCollapse := true;

    if tvGroups.selected <> nil then begin
        AllowCollapse := not (tvGroups.Selected = node);
    end;}
end;

procedure TFrmPermanent.tvGroupsDragOver(Sender, Source: TObject; X, Y: Integer;
  State: TDragState; var Accept: Boolean);
var tn : TTreeNode;
begin
    tn := tvGroups.GetNodeAt(x,y);
    Accept := tn <> nil;
    Accept := Accept and (tvgroups.Selected <> nil); //
    Accept := Accept and (tvgroups.selected.parent <> nil); // root nodes don't move
    Accept := Accept and (tn.Text <> NEW_GROUP_TEXT);

    accepteddrag := Accept;
    if accept then begin


    end;
end;

procedure TFrmPermanent.tvGroupsEndDrag(Sender, Target: TObject; X, Y: Integer);
var drag, drop : TTreeNode;
    dragi, dropi : integer;
    group, prevgroup : string;
    ci : TClipItem;
    name, text : string;
    hotkey : THotkeyData;
    usekeys : boolean;
begin
    if not self.accepteddrag  then EXIT;

    drop := tvGroups.GetNodeAt(x,y);
    drag := tvGroups.Selected;
    //showmessage('selected:' + tvGroups.Selected.Text + ' drop:' + drop.text);

    if (drop <> nil) then begin

        if (drop.Parent <> nil) then begin
            // move item within group

            dragi := drag.Index;
            dropi := drop.Index;
            group := drop.Parent.text;
            //showmessage('drag: ' + IntToStr(drag.Index) + ' drop:' + IntToStr(drop.Index));
            self.PermFolderPush;
            self.SetPermanentPath(group);
            ItemNameList.Move(dragi, dropi);
            ItemDataList.Move(dragi, dropi);
            ItemHotKeyList.Move(dragi, dropi);
            self.SavePermanentItemsGroup;
            self.PermFolderPop;

            //self.SetPermanentPath(group);
            self.RefreshPermanentGroups(false);
            self.SelectTreeItem(group, dropi);
        end else begin
            // move item to new group
            
            dragi := drag.Index;
            group := drop.Text;
            prevgroup := drag.Parent.Text;
            if group = prevgroup then EXIT;
            
            ci := nil;
            self.PermFolderPush;
            self.SetPermanentPath(drag.Parent.text);
                name := ItemNameList[dragi];
                text := ItemDataList[dragi];
                if self.IsComplexItem(text)  then
                    ci := self.GetComplexItem(text);
                hotkey := self.FromString(ItemHotKeyList[dragi]);
                usekeys := lowercase(leftstr(text,6)) = '[keys]';
            self.PermFolderPop;

            self.SaveNewItem(
                group,name,text,hotkey,usekeys,ci
            );

            self.PermFolderPush;
                self.SetPermanentPath(group);
                dropi := ItemNameList.Count - 1;
            self.PermFolderPop ;

            self.PermFolderPush;
                self.SetPermanentPath(prevgroup);
                ItemNameList.Delete(dragi);
                ItemDataList.Delete(dragi);
                ItemHotKeyList.Delete(dragi);
                // TODO - make sure lbItemNames was just included by error
                //lbItemNames.Items.Delete(dragi);
                self.SavePermanentItemsGroup;
            //self.SetPermanentPath(group);
            self.PermFolderPop;
            self.RefreshPermanentGroups(false);
            self.SelectTreeItem(group, dropi);
        end;
    end;

    
end;

procedure TFrmPermanent.tvGroupsExpanding(Sender: TObject; Node: TTreeNode;
  var AllowExpansion: Boolean);
var i : integer;
    tn : TTreeNode;
begin
    AllowExpansion := true;

    //
    // close all other groups except the selected one
    //
    for i  := 0 to tvgroups.Items.count - 1 do begin
        tn := TvGroups.Items[i];

        if (tn.Parent = nil) and
        (tn <> node) then tn.Collapse(true);

    end;

end;

procedure TFrmPermanent.tvGroupsMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var tn : TTreeNode;
begin
    tn := tvGroups.GetNodeAt(x,y);
    if tn <> nil then begin
        //tvGroups.Selected := tn;
{        if tn.Parent = nil then begin
            //tn.Expand(true);
            //tn.selected := true;
        end;
        if tn.Text = NEW_GROUP_TEXT then begin
            self.btnAddPIGClick(self);
            self.tsEditGroupsShow(self);
        end;
}
    end;
end;



procedure TFrmPermanent.ShowSelectedTreeItem;
var tn : TTreeNode;
    //i : integer;
    txt : string;
    ci : TClipItem;
  wc: TWideChar;
  r : Trect;
  hotkey : THotkeyData;
begin

    //tn := tvGroups.GetNodeAt(x,y);
    tn := tvGroups.selected;
    if tn <> nil then begin
        reText.Visible := false;
        imgPic.Visible := false;
        memText.Visible := false;
        btnEditItem.Enabled := false;
        btnNewItem.enabled := false;
        lblhotkey.Caption := EMPTY_HOTKEY_NAME;
        pnltext.Visible := false;
        pnlhotkey.Visible := false;

        if tn.Parent <> nil then begin
            pnltext.Visible := true;
            pnlhotkey.Visible := true;

            self.PermFolderPush;
            self.SetPermanentPath(tn.Parent.text);
            self.SelectedGroup := tn.Parent.text;
            self.SelectedIndex := self.GetTextIndex(tn.text);

            hotkey := self.FromString(ItemHotKeyList[self.selectedindex]);
            lblHotkey.Caption := hotkey.name;
            txt := self.GetItemText(self.SelectedIndex);
            btnEditItem.Enabled := true;

            ci := nil;
            if IsComplexItem(txt) then begin
                ci := self.GetComplexItem(txt);
                if (ci.GetFormat = Windows.CF_DIB) then begin
                    imgPic.Visible := true;
                    ci.GetDIB(imgPic.Picture);
                end else if (ci.GetFormat = unitmisc.GetCF_RICHTEXT)
                    or (ci.GetFormat = unitmisc.GetCF_HTML) then begin
                    reText.Visible := true;
                    reText.PlainText := false;
                    reText.Lines.Clear;
                    ci.GetRichText(txt);
                    reText.Text := txt;
                end else if (ci.GetFormat = Windows.CF_UNICODETEXT) then begin
                    imgPic.Visible := true;
                    wc := TWideChar.Create;
                    wc.append(ci.GetHandle, ci.GetDataSize);


                    imgPic.Canvas.textflags := imgPic.Canvas.textflags or DT_WORDBREAK and (not DT_SINGLELINE);
                    r := imgPic.ClientRect;
                    FillRect(imgPic.Canvas.handle,r,imgPic.Canvas.brush.handle);
                    DrawTextW(
                        imgPic.Canvas.handle, wc.memory, (wc.size-2) div 2,
                        r, DT_WORDBREAK
                    );

                    //lblClipType.Caption := lblClipType.Caption + ' - item contents read only ';

                    myfree(wc);
                end;
            end else begin
                memText.Visible := true;
                memText.Text := txt;
            end;

            MyFree(ci);


            self.PermFolderPop;
        end else begin
            btnNewItem.Enabled := true;
        end;
    end;

end;

procedure TFrmPermanent.tvGroupsMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var tn : TTreeNode;
    i : integer;
    mi : TMenuItem;
    s, cur : string;
begin
    tn := tvgroups.GetNodeAt(x,y);
    if tn <> nil then begin
        if tn.Text = NEW_GROUP_TEXT then EXIT;
        tvgroups.Selected := tn;
    end;

    if tvgroups.selected.text = NEW_GROUP_TEXT  then EXIT;
    self.ShowSelectedTreeItem;


    if button = mbRight then begin
        miMoveto.Enabled := false;
        if tvgroups.Selected.Parent = nil then begin
            midelete.Caption := 'Delete Group';



        end else begin
            miDelete.Caption := 'Delete Item';
            miMoveto.Enabled := true;
            miMoveto.Clear;
            cur := tvgroups.selected.parent.text;

            for i := 0 to FrmPermanent.PermFoldersGetCount - 1 do begin
                s := FrmPermanent.PermFoldersGetItem(i);
                if cur <> s then begin
                    mi := pmTree.CreateMenuItem;
                    mi.Caption := s;
                    mi.OnClick := miMovetoItemClick;
                    mimoveto.Add(mi);

                end;
            end;
        end;

        pmTree.Popup(mouse.CursorPos.X, mouse.CursorPos.Y);
    end;

end;


procedure TFrmPermanent.SaveNewItem(group, name, text: string;
  hotkey: THotKeyData; usekeys: boolean; ci : TClipItem = nil);
begin
    if usekeys and (pos('[keys]', lowercase(text)) <> 0) then begin
        text := rightstr(text, length(text)- length('[keys]'));
    end;

    self.SaveItem(group,-1,name,text,hotkey,usekeys, ci);
end;


procedure TFrmPermanent.SaveItem(group: string; index: integer; name,
  text: string; hotkey: THotKeyData; usekeys: boolean; ci : TClipItem = nil);
var asnew : boolean;
    i : integer;
    hkd2 : THotkeyData;

begin
    self.PermFolderPush;
    self.SetPermanentPath(group);
    asnew := (index = -1);

    // create complex clip in cache, if needed
    if (ci <> nil) then begin
        i := self.GetNextComplexIndex;

        while FileExists(
            clipboard.GetFilename(self.GetPermanentPathFull, i)
        ) do
            inc(i);

        ci.SaveToFile(self.GetPermanentPathFull, i);
        text := '[FILE=' + IntToStr(i) + ']';
    end;


    // create new or save name and text
    if (asnew) then begin
        index := ItemNameList.count;

        // Name must exist, name must be unique for a new item
        //
        if name = '' then name := 'Default';
        i := 1;
        if ItemNameList.IndexOf(name) <> -1 then begin
            while ItemNameList.IndexOf(name + IntToStr(i) ) <> -1 do begin
                inc(i);
            end;
            name := name + IntToStr(i);
        end;

        ItemNameList.Add(name);
        ItemDataList.Add(text);
        ItemHotKeyList.Add('');
    end else begin
        ItemNameList[index] := name;
        ItemDataList[index] := text;
    end;

    // handle keystroke flag for plain text items
    if usekeys and (ci = nil) then begin
        ItemDataList[index] := '[KEYS]' + text;
    end;



    hkd2 := self.FromString(ItemHotKeyList[index]);
    if ToString(hkd2) <> ToString(hotkey) then begin
        if (hotkey.name <> EMPTY_HOTKEY_NAME) then begin
            ItemHotKeyList[index] := self.ToString(hotkey);
        end else begin
            ItemHotKeyList[index] := '';
        end;
    end;


    self.SavePermanentItemsGroup;
    self.PermFolderPop;
    self.RefreshHotkeys;

end;



procedure TFrmPermanent.RefreshPermanentGroups(selectCurrentGroup: boolean);
var i, j : integer;
    tn, subtn : TTreeNode;
begin
    tvGroups.Items.clear;
    self.PermFolderPush;
    for i := 0 to FolderList.Count - 1 do begin
        tn := tvGroups.items.Add(nil, folderlist[i]);
        tn.ImageIndex := 9;
        tn.SelectedIndex := tn.ImageIndex;

        self.SetPermanentPath(folderlist[i]);
        for j := 0 to self.GetCount  - 1 do begin
            subtn := tvGroups.Items.AddChild(tn, self.GetItemName(j));
            subtn.ImageIndex := 3;
            subtn.SelectedIndex := subtn.ImageIndex;

        end;


        if (lowercase(tn.Text) = lowercase(CurrentPermPath)) and
            selectCurrentGroup then begin
            tn.Selected := true;
            tn.Expand(true);
            tvGroups.Selected := tn;
        end else begin
            tn.Selected := false;
        end;
    end;
    self.PermFolderPop;


    tn := tvGroups.items.add(nil, NEW_GROUP_TEXT);
    tn.ImageIndex := 8;
    tn.SelectedIndex := tn.ImageIndex;

    btnNewItem.Enabled := false;
    btnEditItem.Enabled := false;

    self.tvGroupsClick(self);
end;


procedure TFrmPermanent.btnNewItemClick(Sender: TObject);
begin
    FrmPermanentEdit.top := self.top;
    FrmPermanentEdit.left := self.left;

    //if self.SelectedGroup = '' then begin
        self.SelectedGroup := tvGroups.Selected.Text;
    //end;

    FrmPermanentEdit.showform(true, self.SelectedGroup, -1);
    if FrmPermanentEdit.ModalResult <> mrCancel then begin
        self.RefreshPermanentGroups(false);
        self.SelectTreeItem(self.SelectedGroup, -1);
    end;
end;


procedure TFrmPermanent.miDeleteClick(Sender: TObject);
var g : string;
    sr : TSearchRec;
    i : integer;
begin
    if tvGroups.Selected.parent = nil then begin
        //
        // Delete Group
        //
        g := IncludeTrailingPathDelimiter(AppPath) + tvGroups.Selected.text;
        if DirectoryExists(g) then begin
            i := findfirst(g + '\*.*', faAnyFile,  sr);
            while i = 0 do begin
                if (sr.Attr and faDirectory) = 0 then
                    DeleteFile( g + '\' + sr.name );
                i := FindNext(sr);
            end;
            FindClose(sr);

            //deletefile( GetDataFilename(dfn_perm0) );
            //deleteFile( GetDataFilename(dfn_perm1) );
            RmDir(g + '\');

            self.UpdateFolderList;
            self.RefreshPermanentGroups(false); 
        end;
    end else begin
        self.PermFolderPush;
            g := tvGroups.Selected.parent.text;
            self.SetPermanentPath(g);
            i := tvgroups.Selected.Index;
            ItemNameList.Delete(i);
            ItemDataList.Delete(i);
            ItemHotKeyList.Delete(i);
            //lbItemNames.Items.Delete(i);
            self.SavePermanentItemsGroup;
        self.PermFolderPop;
        self.RefreshPermanentGroups(false);
        self.SelectTreeItem(g,-1);
    end;
end;

procedure TFrmPermanent.miMovetoItemClick(Sender: TObject);
var mi : TMenuITem;
    i : integer;
    g, newgroup, name, text : string;
    ci : TClipItem;
    hotkey : THotkeydata;
    usekeys : boolean;

begin
    mi := TMenuItem(sender);
    newgroup := mi.Caption;


    // gather data from selected item
    self.PermFolderPush;
        g := tvGroups.Selected.parent.text;
        self.SetPermanentPath(g);
        i := tvgroups.Selected.Index;
        name := ItemNameList[i];
        text := ItemDataList[i];
        if self.IsComplexItem(text)  then
            ci := self.GetComplexItem(text);
        hotkey := self.FromString(ItemHotKeyList[i]);
        usekeys := lowercase(leftstr(text,6)) = '[keys]';
    self.PermFolderPop;

    // save in new group
    self.SaveNewItem(
        newgroup,name,text,hotkey,usekeys,ci
    );

    // delete item
    self.PermFolderPush;
        self.SetPermanentPath(g);
        ItemNameList.Delete(i);
        ItemDataList.Delete(i);
        ItemHotKeyList.Delete(i);
        self.SavePermanentItemsGroup;
    self.PermFolderPop;
    
    // select new location
        self.SetPermanentPath(g);
        self.RefreshPermanentGroups(true);



end;

end.
