unit UnitFrmSearch;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls, ComCtrls, Contnrs;

type
  TfrmSearch = class(TForm)
    txtFind: TLabeledEdit;
    lvResults: TListView;
    timTypePause: TTimer;
    StatusBar1: TStatusBar;
    procedure FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure FormShow(Sender: TObject);
    procedure txtFindKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure timTypePauseTimer(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormHide(Sender: TObject);
    procedure lvResultsClick(Sender: TObject);
    procedure lvResultsMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure lvResultsKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure lvResultsDrawItem(Sender: TCustomListView; Item: TListItem;
      Rect: TRect; State: TOwnerDrawState);
  private
    { Private declarations }
        LastFind : string;
        TextToPaste : string;
        ObjectList : TObjectList;
  public
    { Public declarations }
        function GetTextToPaste : string;
  end;

type
    TSearchItems = class(TThread)
private
    fabort : boolean;
protected
    procedure Execute; override;
public
    procedure Abort;
end;

var
  frmSearch: TfrmSearch;

implementation

uses UnitFrmPermanentNew, UnitClipQueue, UnitPaste, UnitTWideChar, UnitMisc,
  UnitMenuItemTagdata, UnitFrmMainPopup, UnitPopupGenerate;

{$R *.dfm}

var src : TSearchItems;

procedure TfrmSearch.FormCreate(Sender: TObject);
begin
    src := TSearchItems.Create(true);
    ObjectList := TObjectList.Create;
end;

procedure TfrmSearch.FormHide(Sender: TObject);
begin
    src.Terminate;
    ObjectList.Clear;
end;

procedure TfrmSearch.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
    if key = VK_ESCAPE  then self.ModalResult := mrCancel;

    case key of
    VK_UP:
        begin
        if lvresults.Selected.Index <> 0 then begin
            lvresults.Selected := lvresults.items[lvresults.Selected.index -1];
        end;
        lvresults.Selected.MakeVisible(false);
        key := 0;
        end;
    VK_DOWN:
        begin
        if lvresults.Selected.Index <> (lvresults.items.count-1) then begin
            lvresults.Selected := lvresults.items[lvresults.Selected.index +1];
        end;
        lvresults.Selected.MakeVisible(false);
        
        key := 0;
        end;
    VK_RETURN:
        begin
            key := 0;
            self.modalresult := mrOK;
            self.TextToPaste := '';
            if lvresults.Selected <> nil then
                self.TextToPaste := lvResults.Selected.caption;
        end;
    VK_NEXT: // page down
        begin
            lvResults.Scroll(0,lvResults.height -50);
            lvResults.Selected := lvResults.TopItem;
        end;
    VK_PRIOR: // page up
        begin
            lvResults.Scroll(0,-lvResults.height +50);
            lvResults.Selected := lvResults.TopItem;            
        end;

    end;
end;

procedure TfrmSearch.FormShow(Sender: TObject);
var h : THandle;
begin
    h := Windows.GetForegroundWindow;
    if h <> self.Handle then begin
        if UnitMisc.ThreadAttach(h) then begin
            Windows.SetForegroundWindow(self.handle);
            UnitMisc.ThreadDetach;
        end;
    end;

    txtFind.SetFocus;
    lvResults.Items.Clear;
    if txtfind.Text <> '' then begin
        src.Abort;
        src.Execute;
    end;
end;

function TfrmSearch.GetTextToPaste: string;
begin
    result := TextToPaste;
end;

procedure TfrmSearch.lvResultsClick(Sender: TObject);
begin
    txtFind.SetFocus;
end;

procedure TfrmSearch.lvResultsDrawItem(Sender: TCustomListView; Item: TListItem;
  Rect: TRect; State: TOwnerDrawState);
var wc : TWideChar;
    ci : TClipItem;
    c : TColor;
    mitd : TMenuItemTagData;
begin

    c := TListView(Sender).Canvas.Pen.Color;
    if Item.Selected  then begin
        TListView(Sender).Canvas.Pen.Color := clHighlight;
        TListView(Sender).canvas.Rectangle(rect);
    end;
    TListView(Sender).Canvas.Pen.Color := c;

    mitd := TMenuItemTagData(ObjectList.items[integer(item.data)]);
    if UnitMisc.UniSupported then begin
        ci := mitd.clip;
        wc := TWideChar.Create;
        if assigned(ci) and (ci.GetFormat = CF_UNICODETEXT) then begin
            wc.Append(ci.GetHandle, ci.GetDataSize);
        end else begin
            wc.Append(ci.GetAsText);
        end;
        Windows.TextOutW(
            TListView(Sender).Canvas.Handle,
            rect.Left, rect.Top, wc.GetMemoryStream.Memory, wc.StrLength
            );
    end else begin
        Windows.TextOut(
            TListView(Sender).Canvas.Handle,
            rect.Left, rect.Top, pchar(mitd.caption) , length(mitd.caption)
            );
    end;

end;

procedure TfrmSearch.lvResultsKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
    txtFind.SetFocus;
end;

procedure TfrmSearch.lvResultsMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
  var li : TListItem;
begin
    li := lvResults.GetItemAt(x,y);
    if li <> nil then begin
        self.TextToPaste := li.caption;
        self.ModalResult := mrOK;
    end;
end;

procedure TfrmSearch.timTypePauseTimer(Sender: TObject);
begin
    timTypePause.Enabled := false;
    if txtFind.text = '' then lvResults.Items.Clear;
    if lastfind <> txtFind.text then begin
        lvResults.Items.Clear;
        //Src.Terminate;
        src.Abort;
        Src.Execute;
    end;

    lastFind := txtFind.text;
end;

procedure TfrmSearch.txtFindKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
    timTypePause.Enabled := false;
    timTypePause.Enabled := true;
end;

{ TSearchItems }

procedure TSearchItems.Abort;
begin
    self.fabort := true;
end;

procedure TSearchItems.Execute;
var i,j, x : integer;
    find, text : string;
    li : TListItem;

    function TestText(sub, source : string) : boolean;
    var res : integer;
        original : string;
    begin
        original := source;
        sub := lowercase(sub);
        source := lowercase(source);

        // favor items with searchstring at the start
        res := pos(sub, source);
        li := nil;
        if (res > 0) and (res < 5) then begin
            li := FrmSearch.lvResults.Items.Insert(0);
        end else if (res > 0) then begin
            li := frmSearch.lvResults.Items.Add;
        end;
        result := false;
        if (li <> nil) then begin
            li.Caption := Original;
            result := true;
        end;
    end;
var mitd : TMenuItemTagData;
begin
    inherited;
    fabort := false;
    frmsearch.StatusBar1.SimpleText := 'Searching ...';
    frmsearch.lvResults.Height := 185;
    frmsearch.AutoSize := false;
    frmsearch.Height := 287;
    frmsearch.AutoSize := true;
    Application.ProcessMessages;

    find := frmSearch.txtFind.Text;
    if find <> '' then begin
        // search both name and contents of Perm Items
        for i := 0 to FrmPermanent.PermFoldersGetCount - 1 do begin
            FrmPermanent.PermFolderPush;
                FrmPermanent.SetPermanentPath( frmPermanent.PermFoldersGetItem(i) );
                for j := 0 to FrmPermanent.GetCount - 1 do begin
                    text := FrmPermanent.GetItemName(j);
                    if not TestText(find, text) then begin
                        text := FrmPermanent.GetItemText(j);
                        if not FrmPermanent.IsComplexItem(text) then begin
                            if TestText(find, text) then begin
                                mitd := TMenuItemTagData.Create;
                                li.Data := Pointer(frmsearch.ObjectList.add(mitd));
                                mitd.caption := text;
                                mitd.itemtype := IT_PERMANENT;
                                mitd.PermanentGroupID := i;
                                mitd.itemindex := j;
                            end;
                        end
                    end else begin
                        li.caption := FrmPermanent.GetItemText(j);
                        mitd := TMenuItemTagData.Create;
                        li.Data := Pointer(frmsearch.ObjectList.add(mitd));
                        mitd.caption := li.caption;
                        mitd.itemtype := IT_PERMANENT;
                        mitd.PermanentGroupID := i;
                        mitd.itemindex := j;
                    end;

                    if self.fabort then BREAK;
                end;
            FrmPermanent.PermFolderPop;
            if self.fabort then EXIT;
        end;

        for i := 0 to ClipQueue.GetQueueCount - 1 do begin
            text := ClipQueue.GetItemText(i);
            if TestText(find, text) then begin
                mitd := TMenuItemTagData.Create;
                li.Data := Pointer(frmsearch.ObjectList.add(mitd));
                mitd.caption := text;
                mitd.clip := ClipQueue.GetClipItem(i);
                mitd.itemtype := IT_MENU;
                mitd.itemindex := i;
            end;
            if self.fabort then EXIT;
        end;

        for i := 0 to TempClipQueue.GetQueueCount  - 1 do begin
            text := TempClipQueue.GetItemText(i);
            if TestText(find, text) then begin
                mitd := TMenuItemTagData.Create;
                li.Data := Pointer(frmsearch.ObjectList.add(mitd));
                mitd.caption := text;
                mitd.itemtype := IT_TEMP;
                mitd.clip := TempClipQueue.GetClipItem(i);
                mitd.itemindex := i;
            end;
            if self.fabort then EXIT;
        end;

        for i := 0 to RemovedQueue.GetQueueCount  - 1 do begin
            text := RemovedQueue.GetItemText(i);
            if TestText(find, text) then begin
                mitd := TMenuItemTagData.Create;
                li.Data := Pointer(frmsearch.ObjectList.add(mitd));
                mitd.caption := text;
                mitd.itemtype := IT_OTHER;
                mitd.clip := RemovedQueue.GetItemClip(i);
                mitd.itemindex := i;
            end;
            if self.fabort then EXIT;
        end;
    end;

    frmsearch.StatusBar1.SimpleText := '';
    Application.ProcessMessages;


    if frmsearch.lvResults.selected = nil then begin
        if frmsearch.lvResults.Items.Count > 0 then begin
            frmsearch.lvResults.selected := frmsearch.lvResults.items[0];
        end;
    end;
    x := frmsearch.lvResults.Canvas.TextHeight('ABCDEFGHIJLKMNONQ');
    while (frmsearch.lvResults.VisibleRowCount < frmsearch.lvResults.Items.Count) and
        (frmsearch.height < 650) do begin
        frmsearch.lvResults.Height := frmsearch.lvResults.Height + X + 2;
        Application.ProcessMessages;
    end;
    frmsearch.lvResults.Height := frmsearch.lvResults.Height + X * 2;

end;

end.
