unit UnitFrmSysTrayMenu;
{
    Purpose:
        This unit is responsible for the system tray icon and it's
        popup menu.
    Updates:
        Updated for option to show popup on double-click

        -------------
        Show the version in the Icon "Hint"
        -------------
        Updated method of modifying the systray icon so it does change
        the systray or move in position.

        Added public method to disable/enable systray icon changing

        Figured out a simple/reliable way to show picture on menu item
        Using UnitClipQueue
}
interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Menus, ExtCtrls, ShellAPI, StdCtrls, ImgList, XPMan, StdActnMenus,
  ActnPopup, XPStyleActnCtrls, UnitFrmDummyUnicodeTooltip, ActnMan;


const MY_WM_TRAYICON = WM_USER + 1; // used to receive tray icon messsages


{type TMYPopupActionBar = class(TPopupActionBar)
    public
        constructor Create(AOwner: TComponent); override;
end;
}
type
  TfrmSysTrayMenu = class(TForm)
    Label1: TLabel;
    ToggleTimer: TTimer;
    ImageList1: TImageList;
    timDisabled: TTimer;
    PopupActionBar1: TPopupActionBar;
    ilPopupIcons: TImageList;
    test1: TMenuItem;
    test21: TMenuItem;
    timCloseTooltip: TTimer;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure timDisabledTimer(Sender: TObject);
    procedure timCloseTooltipTimer(Sender: TObject);
  private
    { Private declarations }

    TrayIcon: TNotifyIconData;          {tray icon info}
    iconPic : TIcon;
    DisableIconChanges : boolean;
    PopupShowing : boolean;
    WM_TASKBAR_CREATE : Cardinal;
    ForceIconRestore : boolean;
    JustSwitched : boolean;
    targetEXE : string;
    CurrentIconIndex : integer;
    DisalbedEXEs : TStringList;
    NoRightClickEXEs : TStringList;
    tooltip : TTooltipWindow;
    notooltip : boolean;

    procedure ExitMenuItemClickEvent(Sender: TObject);
    //procedure FlushHistoryMenuItemClickEvent(sender : TObject);


    function GetIconTooltipText : string;


    procedure AboutClickEvent(sender : TObject);
    procedure SetIconByIndex(index : longint);
    procedure CreateTrayIcon;
  public
    TrayPopup : TPopupMenu;

    procedure EditHistoryMenuItemClickEvent(sender : TObject);
     procedure EditClipboardMenuItemClickEvent(sender : TObject);

    procedure ConfigureMenuItemClickEvent(Sender : TObject);
    procedure RefreshMenuItemClickEvent(Sender : TObject);
    procedure FlushTextNonTextClickEvent(Sender : TObject);
    procedure FlushEverything(Sender : TObject);
    procedure FlushClipboard(Sender: TObject);
    procedure ClipboardBarShow(Sender : TObject);
    procedure ClipboardBarHide(Sender : Tobject);


    procedure DisabledEXEClickEvent(Sender : TObject);
        procedure NoRightClickEvent(Sender : TObject);
      procedure RemovedItemsMenuItemClickEvent(sender : Tobject);

    procedure SwitchMenuItemClickEvent(sender : TOBject);
    procedure MethodMenuItemClickEvent(Sender : TObject);
    procedure SwitchPermFolder(sender : TObject);
    function IsDisabledEXE(exename : string) : boolean;
    function IsNoRightClickEXE(exename : string) : boolean;
    procedure ToggletDisabledEXE(exename : string);
    procedure ToggleNoRightClickEXE(exename : string);
    { Public declarations }

        procedure WMTrayIcon(var Msg: TMessage); message MY_WM_TRAYICON;
        procedure WndProc(var Msg: TMessage); override; // trap new icon message

        procedure SetIconNormal;
        procedure SetIconIgnore;
        procedure SetIconTriggered;
        procedure SetDisableIconChanges(value : boolean);
        function GetJustSwitched : boolean;


        procedure ShowPopup(inmenu : TMenuItem = nil);
  end;

var
  frmSysTrayMenu: TfrmSysTrayMenu;

implementation

uses  UnitFrmConfig, UnitFrmClipboardManager,clipbrd,
  UnitFrmMainPopup,UnitPopupGenerate, UnitClipQueue, UnitFrmAbout, UnitFrmRemoved,
  UnitKeyboardQuery, UnitFrmPermanentNew, UnitMisc, UnitOtherQueue,
  UnitPaste, UnitMenuItemTagdata, UnitFrmEditItem , UnitFrmDummyClipboardBar, ActnMenus
  ;
{$R *.dfm}

CONST INDEX_IGNORE = 0;
        INDEX_NORMAL = 2;
        INDEX_TRIGGERED = 1;


{////////////////////}
{//Public Interface//}
{////////////////////}

procedure TfrmSysTrayMenu.ToggleNoRightClickEXE(exename: string);
begin
    self.targetEXE := exename;
    self.NoRightClickEvent(self);
end;

procedure TfrmSysTrayMenu.ToggletDisabledEXE(exename: string);
begin
    self.targetEXE := exename;
    self.DisabledEXEClickEvent(self);
end;

procedure TFrmSysTrayMenu.SetDisableIconChanges(value: boolean);
begin
   self.DisableIconChanges := value;
end;

procedure TFrmSysTrayMenu.SetIconNormal;
var s : string;
begin
    ToggleTimer.Enabled := false;

    {Do I really need disable updates if the icon is normal?}
    //if (not self.DisableIconChanges) or (self.ForceIconRestore) then begin
        self.CurrentIconIndex := INDEX_NORMAL;
        self.ForceIconRestore := false;
        TrayIcon.cbSize := SizeOf(TrayIcon);
        TrayIcon.Wnd := Self.Handle;
        TrayIcon.uID := 0;
        TrayIcon.uFlags := NIF_ICON or NIF_TIP or NIF_MESSAGE;
        TrayIcon.uCallbackMessage := MY_WM_TRAYICON;
        TrayIcon.hIcon := Application.Icon.Handle;
        //TrayIcon.szTip := 'ArsClip';
        s := GetIconTooltipText;
        move(s[1], TrayIcon.szTip, length(s));


        ShellAPI.Shell_notifyIcon(NIM_MODIFY, @TrayIcon);
    //end;
end;

procedure TFrmSysTrayMenu.SetIconIgnore;
begin
    self.SetIconByIndex(INDEX_IGNORE);
end;

procedure TFrmSysTrayMenu.SetIconTriggered;
begin
    self.SetIconByIndex(INDEX_TRIGGERED);
end;

procedure TFrmSysTrayMenu.SetIconByIndex(index : longint);
var s : string;
begin
    ToggleTimer.Enabled := true;

    if (not self.DisableIconChanges) then begin
        self.CurrentIconIndex := index;
        ImageList1.GetIcon(index, IconPic);
        TrayIcon.cbSize := SizeOf(TrayIcon);
        TrayIcon.Wnd := Self.Handle;
        TrayIcon.uID := 0;
        TrayIcon.uFlags := NIF_ICON or NIF_TIP or NIF_MESSAGE;
        TrayIcon.uCallbackMessage := MY_WM_TRAYICON;

        TrayIcon.hIcon :=  IconPic.Handle;
        //TrayIcon.szTip := 'ArsClip';
        s := GetIconTooltipText;
        move(s[1], TrayIcon.szTip, length(s));

        ShellAPI.Shell_notifyIcon(NIM_MODIFY, @TrayIcon);
    end;
end;


{
--========================
-- // Create/Destroy   //
--========================
}
const DISABLED_FILE = 'disabled.ini';
const NORIGHTCLICK_FILE = 'noright.ini';
procedure TfrmSysTrayMenu.FormCreate(Sender: TObject);
var s : string;
begin
    UnitMisc.AppendLog('SystemTrayIcon - creating');

    if assigned(FrmConfig)  and FrmConfig.GetOldTrayPopup then begin
        TrayPopup := TPopupMenu.Create(self);
    end else begin
        TrayPopup := TPopupActionBar.Create(self);
    end;
    iconpic := TIcon.Create;


    self.CreateTrayIcon;
    WM_TASKBAR_CREATE := RegisterWindowMessage('TaskbarCreated');

    self.DisalbedEXEs := TStringList.Create;
    self.NoRightClickEXEs := TStringList.Create;
    s := UnitMisc.GetAppPath;
    s := s + DISABLED_FILE;
    if FileExists(s) then begin
        self.DisalbedEXEs.LoadFromFile(s);
    end;
    s := UnitMisc.GetAppPath;
    s := s + NORIGHTCLICK_FILE;
    if FileExists(s) then begin
        self.NoRightClickEXEs.LoadFromFile(s);
    end;


    tooltip := TTooltipWindow.Create;
end;

procedure TfrmSysTrayMenu.CreateTrayIcon;
var s : string;
begin
    //
    // create the initial icon
    // NIM_MODIFY will be performed for updates
    //
    TrayIcon.cbSize := SizeOf(TrayIcon);
    TrayIcon.Wnd := Self.Handle;
    TrayIcon.uID := 0;
    TrayIcon.uFlags := NIF_ICON or NIF_TIP or NIF_MESSAGE;
    TrayIcon.uCallbackMessage := MY_WM_TRAYICON;
    TrayIcon.hIcon := Application.Icon.Handle;
    s := GetIconTooltipText;
    move(s[1], TrayIcon.szTip, length(s));

    ShellAPI.Shell_notifyIcon(NIM_ADD, @TrayIcon);

end;


procedure TfrmSysTrayMenu.FormDestroy(Sender: TObject);
var s : string;
begin
    //
    // Remove tray icon
    //
    ShellAPI.Shell_notifyIcon(NIM_DELETE, @TrayIcon);

    TrayPopup.Items.Clear;
    MyFree(TrayPopup);
    MyFree(iconPic);

    s := UnitMisc.GetAppPath;
    s := s + DISABLED_FILE;
    self.DisalbedEXEs.SaveToFile(s);
    s := UnitMisc.GetAppPath;
    s := s + NORIGHTCLICK_FILE;
    self.NoRightClickEXEs.SaveToFile(s);
    MyFree(DisalbedEXEs);
    MyFree(NoRightClickEXEs);
end;






{
--========================
-- // Trayicon messages //
--========================
                                          r
Description: On a right click on the tray icon,
    present a popup menu that allows the user to

close the program
run configuration
run permanent items
switch permanent items
flush the clipboard history

}


procedure TfrmSysTrayMenu.WndProc(var Msg: TMessage);
begin
    // force a restore if explorer crashes
    if msg.Msg = self.WM_TASKBAR_CREATE  then begin
        self.ForceIconRestore := true;
        self.CreateTrayIcon;
        EXIT;
    end else begin
        inherited;
    end;

    {
    // the tray icon never receives many messages, like below
    case msg.msg of
    WM_MOUSEHOVER:
        begin
            showmessage('hover');
        end;
    else
        inherited;
    end;
    }
end;


procedure TFrmSysTrayMenu.WMTrayIcon(var Msg: TMessage);
var ci : TClipItem;
begin
    //
    // show Popup menu on a right click
    // show configuration form on a left double click
    // show/hide debug window on right double click
    //

    case msg.LParam of
    WM_LBUTTONDOWN : begin
        if (FrmConfig.cbPopupSingleclick.checked) then begin
            FrmMainPopup.ShowOnNextWindow;
        end else if (FrmConfig.cbPopupSingleclickTray.checked) then begin
            FrmMainPopup.ShowOnSystemTray;
        end else begin
            ci := tclipitem.Create;
            ci.GetClipboardItem(0);

            tooltip.ShowTooltip(
                ci,
                point(mouse.CursorPos.X+19,mouse.cursorpos.Y),
                self.GetIconTooltipText,
                false
            );
            self.timCloseTooltip.Enabled := true;


            myfree(ci);
        end;
    end;

    WM_MOUSEMOVE: begin
        if (tooltip.IsShowing) and (not self.notooltip) then begin
            self.timCloseTooltip.Enabled := false;
            self.timCloseTooltip.Enabled := true;
        end;
    end;
    WM_Rbuttondown : begin
        self.notooltip := true;
        tooltip.CloseTooltip; 
        self.ShowPopup;
        self.notooltip := false;
    end;
    WM_LBUTTONDBLCLK : begin
        self.notooltip := true;
        tooltip.CloseTooltip;

        if (FrmConfig.cbPopupDoubleclick.checked) then begin
            FrmMainPopup.ShowOnNextWindow;
        end else if (FrmConfig.cbPopupDoubleclickTray.checked) then begin
            FrmMainPopup.ShowOnSystemTray;
        end else begin
            frmconfig.show();
        end;
        self.notooltip := false;

    end;
    WM_RBUTTONDBLCLK : begin
        frmMainPopup.Visible := frmMainPopup.Visible xor true;
    end;
    end;


end;




procedure TFrmSysTrayMenu.ShowPopup(inmenu : TMenuItem);
var CursorPos : TPoint;
    menu: TMenuItem;
    submenu : TMenuItem;
    //TrayPopup : TPopupMenu;
    i : integer;
    s : string;
    {procedure ImageToBitmap(img : TImage; bmp : TBitmap);
    begin
        bmp.pixelformat := pf24bit;
        bmp.Width := img.Width;
        bmp.Height := img.Height;
        bmp.Canvas.draw(0,0, img.Picture.Graphic);
    end;}
    procedure AddToPopup(item : TMenuItem);
    begin
        if (inmenu = nil) then begin
            TrayPopup.Items.Add(item);
        end else begin
            inmenu.Add(item);
        end;
    end;
begin
    if (self.PopupShowing) then EXIT;
    self.PopupShowing := true;


    {
    |=================|
    | Configure       |
    | Assign method to [program] |>|
    |-----------------|
    | Permenant Clips |
    | Switch to... |> |
    |-----------------|
    | Flush History   |
    | Edit History    |
    | Edit Clipboard..|
    | Removed Items...|
    |-----------------|
    | About           |
    | Exit            |
    |=================|
    }
    UnitMisc.AppendLog('WMTrayIcon right button ');

    if (inmenu = nil) then begin
        Windows.GetCursorPos(CursorPos);
        //MyFree(TrayPopup);

        //TrayPopup := TPopupActionBar.Create(nil);
        //TrayPopup := TPopupMenu.Create(nil);
        TrayPopup.Items.Clear;
        TrayPopup.AutoHotkeys := maAutomatic;
    end;

    MyFree(TrayPopup);
    if FrmConfig.GetOldTrayPopup then begin
        TrayPopup := TPopupMenu.Create(self);
    end else begin
        TrayPopup := TPopupActionBar.Create(self);
    end;

    traypopup.Images := ilPopupIcons;
    // for dynamically created popup, the image list isn't set yet
    // and must be forced

    menu := TMenuItem.Create(TrayPopup);
    menu.caption := '&Configure ...';
    menu.OnClick := self.ConfigureMenuItemClickEvent;
    //menu.Bitmap := FrmMainPopup.imgA.Picture.Bitmap;
    menu.ImageIndex := 0;
    menu.Bitmap := nil;
    //menu.hint := IntToStr(Integer(@FrmMainPopup.imgA));
    AddToPopup(menu);

    if not frmClipboardManager.IsNewClipboardAPI then begin
        menu := TMenuItem.Create(TrayPopup);
        menu.caption := '&Refresh Clipboard Monitoring';
        menu.OnClick := self.RefreshMenuItemClickEvent;
        menu.ImageIndex := -1;
        //menu.Bitmap := FrmMainPopup.imgA.Picture.Bitmap;
        AddToPopup(menu);
    end;


    {
    submenu := TMenuItem.Create(menu);
    submenu.Caption := '&6 Disable Monitoring';
    submenu.Tag := Integer(PASTE_DISABLED);
    submenu.OnClick := self.MethodMenuItemClickEvent;
    submenu.Checked := false;
    menu.Add(submenu);
    }
    //AddToPopup(menu);
    AddToPopup(NewLine());


    menu := TMenuItem.Create(TrayPopup);
    menu.Caption := 'Permanent Clips ...        ';
    menu.OnClick := self.SwitchMenuItemClickEvent;
    menu.Bitmap := FrmMainPopup.imgRightArrow.Picture.Bitmap;
    menu.ImageIndex := 1;
    //menu.hint := IntToStr(Integer(@FrmMainPopup.imgRightArrow));
    AddToPopup(menu);

    if (inmenu = nil) then begin
        // switch to sub menu (checkmark current item)
        menu := TMenuItem.Create(TrayPopup);
        menu.Caption := 'Switch to Group     ';
        s := frmPermanent.GetPermanentPath;
        for i := 0 to frmPermanent.PermFoldersGetCount - 1 do begin
            submenu := TMenuItem.Create(menu);
            submenu.hint := frmPermanent.PermFoldersGetItem(i);
            submenu.caption := submenu.hint;
            if (submenu.Hint = s) then begin
                submenu.Checked := true;
            end;
            submenu.OnClick := self.SwitchPermFolder;
            menu.Insert(i, submenu)
        end;
        AddToPopup(menu);
    end;


    menu := TMenuItem.Create(TrayPopup);
    menu.Caption := 'Removed Clips ...';
    menu.Bitmap := FrmMainPopup.imgRemoved.Picture.Bitmap;
    //menu.hint := IntToStr(Integer(@FrmMainPopup.imgRemoved));
    menu.OnClick := self.RemovedItemsMenuItemClickEvent;
    menu.ImageIndex := 5;
    AddToPopup(menu);


    AddToPopup(NewLine());

    submenu := TMenuItem.Create(TrayPopup);
    submenu.Caption := 'Flush';
    submenu.OnClick := nil;
    submenu.Bitmap := FrmMainPopup.imgFlush.Picture.Bitmap;
    submenu.ImageIndex := 2;



    menu := TMenuItem.Create(TrayPopup);
    menu.Caption := 'Flush Clipboard';
    menu.Bitmap := FrmMainPopup.imgFlush.Picture.Bitmap;
    menu.ImageIndex := 2;

    //menu.hint := IntToStr(Integer(@FrmMainPopup.imgFlush));
    menu.OnClick := Self.FlushClipboard;
    submenu.Add(menu);
    submenu.Add(NewLine());


    menu := TMenuItem.Create(TrayPopup);
    menu.Caption := 'Flush Popup Text';
    menu.Bitmap := FrmMainPopup.imgFlush.Picture.Bitmap;
    //menu.hint := IntToStr(Integer(@FrmMainPopup.imgFlush));
    menu.OnClick := FrmMainPopup.FlushTextItemsClickEventNoPop;
    submenu.Add(menu);



    menu := TMenuItem.Create(TrayPopup);
    menu.Caption := 'Flush Popup Non-Text';
    menu.Bitmap := FrmMainPopup.imgFlush.Picture.Bitmap;
    //menu.hint := IntToStr(Integer(@FrmMainPopup.imgFlush));
    menu.OnClick := FrmMainPopup.FlushOtherItemsClickEventNoPop;
    submenu.Add(menu);

    menu := TMenuItem.Create(TrayPopup);
    menu.Caption := 'Flush Popup Text && Non-Text';
    menu.Bitmap := FrmMainPopup.imgFlush.Picture.Bitmap;
    //menu.hint := IntToStr(Integer(@FrmMainPopup.imgFlush));
    menu.OnClick := self.FlushTextNonTextClickEvent;
    submenu.Add(menu);
    submenu.Add(NewLine());

    menu := TMenuItem.Create(TrayPopup);
    menu.Caption := 'Flush Everything (including removed) ';
    menu.Bitmap := FrmMainPopup.imgFlush.Picture.Bitmap;
    //menu.hint := IntToStr(Integer(@FrmMainPopup.imgFlush));
    menu.OnClick := self.FlushEverything;
    menu.ImageIndex := 2;

    submenu.Add(menu);


    AddToPopup(submenu);
    //AddToPopup(NewLine());


    menu := TMenuItem.Create(TrayPopup);
    menu.Caption := 'Edit History ...';
    menu.Bitmap := FrmMainPopup.imgEdit.Picture.Bitmap;
    //menu.hint := IntToStr(Integer(@FrmMainPopup.imgEdit));
    menu.OnClick := self.EditHistoryMenuItemClickEvent;
    menu.ImageIndex := 3;

    AddToPopup(menu);

    menu := TMenuItem.Create(TrayPopup);
    menu.Caption := 'Edit Clipboard ...';
    menu.Bitmap := FrmMainPopup.imgPaste.Picture.Bitmap;
    //menu.hint := IntToStr(Integer(@FrmMainPopup.imgEdit));
    menu.OnClick := self.EditClipboardMenuItemClickEvent;
    menu.ImageIndex := 4;

    AddToPopup(menu);


    menu := TMenuItem.Create(TrayPopup);
    menu.Caption := 'Clipboard Bar';
    menu.OnClick := nil;
    menu.Bitmap := FrmMainPopup.imgPaste.Picture.Bitmap;
    menu.ImageIndex := 6;
    submenu := TMenuItem.Create(TrayPopup);
    submenu.Caption := 'Show';
    if assigned(frmDummyClipboardBar) then begin
        submenu.Checked := frmDummyClipboardBar.IsVisible;
    end else begin
        submenu.checked := false;
    end;
    submenu.OnClick := self.ClipboardBarShow;
    menu.Add(submenu);
    submenu := TMenuItem.Create(TrayPopup);
    submenu.Caption := 'Hide';
    if assigned(frmDummyClipboardBar) then begin
        submenu.Checked := not frmDummyClipboardBar.IsVisible;
    end else begin
        submenu.checked := true;
    end;
    submenu.OnClick := self.ClipboardBarHide;
    menu.Add(submenu);
    AddToPopup(menu);




    AddToPopup(NewLine());
    menu := TMenuItem.Create(TrayPopup);
    menu.caption := 'About';
    menu.OnClick := self.AboutClickEvent;
    AddToPopup(menu);


    menu := TMenuItem.Create(TrayPopup);
    menu.caption := 'Exit ArsClip';
    menu.OnClick := self.ExitMenuItemClickEvent;
    AddToPopup(menu);





    //UnitMisc.MySetTheme( GetMenu( ); 
    if (inmenu = nil) then begin
        UnitMisc.AppendLog('>>tray popup');
        Windows.SetForegroundWindow(frmMainPopup.handle);


        try
            TrayPopup.Popup(CursorPos.x+2, CursorPos.y-4);

        except
            // automagically catch and show the old popup on a crash

            if not FrmConfig.GetOldTrayPopup then begin
                FrmConfig.ForceOldTrayPopup;
                self.ShowPopup(inmenu);
            end;

        end;
        Windows.PostMessage(frmMainPopup.Handle, WM_NULL, 0, 0);
        UnitMisc.AppendLog('>>tray popup done');
    end;


    self.PopupShowing := false;
end;


{
  // menu item clicks
}


procedure TFrmSysTrayMenu.RefreshMenuItemClickEvent(Sender : TObject);
begin
    frmClipboardManager.RefreshClipboardMonitor;
end;
procedure TFrmSysTrayMenu.ExitMenuItemClickEvent(Sender: TObject);
begin
    // End the program nicely
    frmMainPopup.Close;
end;
procedure TFrmSysTrayMenu.SwitchMenuItemClickEvent(sender : TOBject);
begin
    Windows.SetForegroundWindow(frmPermanent.Handle);
    frmPermanent.Show;
end;
procedure TfrmSysTrayMenu.ClipboardBarHide(Sender: Tobject);
begin
    if assigned(frmdummyclipboardbar) then begin
        frmDummyClipboardBar.SetVisible(false);
    end;
end;

procedure TfrmSysTrayMenu.ClipboardBarShow(Sender: TObject);
begin
    if assigned(frmdummyclipboardbar) then begin
        frmDummyClipboardBar.SetVisible(true);
    end;
end;

procedure TFrmSysTrayMenu.ConfigureMenuItemClickEvent(Sender : TObject);
begin
    Windows.SetForegroundWindow(frmConfig.Handle);
    frmConfig.Show;
end;
procedure TFrmSysTrayMenu.SwitchPermFolder(sender : TObject);
begin
    with sender as TMenuItem do begin
        frmPermanent.SetPermanentPath( hint );
    end;

    self.JustSwitched := true;
end;

procedure TfrmSysTrayMenu.timCloseTooltipTimer(Sender: TObject);
begin
    tooltip.CloseTooltip;
    timCloseTooltip.Enabled := false;
end;

procedure TfrmSysTrayMenu.timDisabledTimer(Sender: TObject);
begin
    if not frmClipboardManager.GetMonitoring  then begin
        self.SetIconIgnore;
        EXIT;
    end;
    if assigned(frmconfig) then
        If (FrmConfig.cbDisableWhenScrollLock.checked) then begin
            if (KeyboardQuery.LockOn(VK_SCROLL)) then begin
                self.SetIconIgnore;
                EXIT;
            end;
        end;
    if self.CurrentIconIndex <> INDEX_NORMAL then begin
        self.SetIconNormal;
    end;
    
end;

procedure TfrmSysTrayMenu.DisabledEXEClickEvent(Sender: TObject);
var s : string;
begin
    s := self.DisalbedEXEs.values[targetEXE];
    if s = '' then begin
        self.DisalbedEXEs.values[targetEXE] := '1';
    end else begin
        self.DisalbedEXEs.values[targetEXE] := '';
    end;

    targetEXE := '';
end;


procedure TfrmSysTrayMenu.NoRightClickEvent(Sender: TObject);
var s : string;
begin
    s := self.NoRightClickEXEs.values[targetEXE];
    if s = '' then begin
        self.NoRightClickEXEs.values[targetEXE] := '1';
    end else begin
        self.NoRightClickEXEs.values[targetEXE] := '';
    end;

    targetEXE := '';

end;


{
procedure TFrmSysTrayMenu.FlushHistoryMenuItemClickEvent(sender : TObject);
begin
    if (Dialogs.MessageDlg('Flush all text and non-text items?',
        mtConfirmation, [mbYes,mbNo],0) = mrYes) then begin
        ClipQueue.ClearQueue;
        OtherQueue.ClearQueue;
    end;
end;
}
procedure TfrmSysTrayMenu.EditClipboardMenuItemClickEvent(sender: TObject);
var p : TPosition;
begin
    FrmEditItem.SetText(clipbrd.Clipboard.AsText, nil);
    //Windows.SetForegroundWindow(FrmEditItem.Handle);

    p := FrmEditItem.Position;
    FrmEditItem.Position := poScreenCenter;
    FrmEditItem.Left := mouse.CursorPos.X;
    FrmEditItem.Top := mouse.cursorpos.y;
    FrmEditItem.ShowModal;
    FrmEditItem.Position := p;
end;

procedure TFrmSysTrayMenu.EditHistoryMenuItemClickEvent(sender : TObject);
begin
    Windows.SetForegroundWindow(frmClipboardManager.Handle);
    frmClipboardManager.Show;
end;

procedure TFrmSysTrayMenu.AboutClickEvent(sender : TObject);
begin
    Windows.SetForegroundWindow(frmAbout.Handle);
    frmAbout.show;
end;


procedure TFrmSysTrayMenu.RemovedItemsMenuItemClickEvent(sender : Tobject);
begin
    Windows.SetForegroundWindow(FrmRemoved.Handle);
    FrmRemoved.Show;
end;


function TfrmSysTrayMenu.GetIconTooltipText: string;
begin
    result := 'ArsClip v' + FrmAbout.lblversion.caption;

    if assigned(frmconfig) then
        result := 'ArsClip v' + FrmAbout.lblversion.caption +
        '  [' + frmconfig.GetHumanReadableHotkey + ']' + #0;
end;

function TfrmSysTrayMenu.GetJustSwitched: boolean;
begin
    result := self.JustSwitched;
    self.JustSwitched := false;
end;

procedure TfrmSysTrayMenu.MethodMenuItemClickEvent(Sender: TObject);
var m : TMenuItem;
    pm : TPasteMethod;
    mitd: TMenuItemTagData;
begin
    m := TMenuItem(sender);


    if self.targetEXE = '' then begin
        self.targetEXE := FrmMainPopup.GetTargetEXE;
        mitd := TagDataList.get(TMenuItem(sender).tag);
        pm := TPasteMethod(mitd.itemindex);
    end else begin
        pm := TPasteMethod(m.tag);
    end;
    Paste.AssignPaste(self.targetEXE, pm);

    self.targetEXE := '';
end;





function TfrmSysTrayMenu.IsDisabledEXE(exename: string): boolean;
begin
    result := self.DisalbedEXEs.Values[exename] <> '';
end;

function TfrmSysTrayMenu.IsNoRightClickEXE(exename: string): boolean;
begin
    result := self.NoRightClickEXEs.Values[exename] <> '';
end;

procedure TfrmSysTrayMenu.FlushTextNonTextClickEvent(Sender: TObject);
begin
    FrmMainPopup.FlushTextItemsClickEventNoPop(self);
    FrmMainPopup.FlushOtherItemsClickEventNoPop(self);
end;

procedure TfrmSysTrayMenu.FlushClipboard(Sender: TObject);
begin
    clipbrd.Clipboard.Clear;
end;

procedure TfrmSysTrayMenu.FlushEverything(Sender: TObject);
begin
    self.FlushTextNonTextClickEvent(self);
    FrmRemoved.bFlushItems.Click;
    TempClipQueue.ClearQueue;
    ClickedClipQueue.ClearQueue;
    self.FlushClipboard(self);
    clipbrd.clipboard.Clear;

    FrmMainPopup.ClearCurrentLast;
end;


{ TMYPopupActionBar }
{
constructor TMYPopupActionBar.Create(AOwner: TComponent);
var mystyle : TXPStyleActionBars;
    colormap : TCustomActionBarColorMap;
begin
    inherited;
    self.PopupMenu := TCustomActionPopupMenuEx.Create(self);
    mystyle := self.PopupMenu.ActionManager.Style;
end;
}
end.
