unit UnitMisc;
{
    Purpose:
        "Put all the orphans in one bed"

    Updates:
        Handle error case for WindowHandleToEXEName
        Routine to save the debug log

}

{ TODO: Further optimize log usage - possibly to file }
interface

uses Windows, SysUtils, StrUtils;

procedure MyFree(var o);
function WindowHandleToEXEName(handle : THandle) : string;
function DupHandle(h : Thandle; var sizeh : cardinal; SizeLimit : boolean = false) : Thandle;
procedure AppendLog( s : string; IncludeLastError : boolean = false);
procedure DumpLog(filename : string);
procedure ApplicationException(Sender: TObject; E: Exception);


procedure FlushLog;
procedure MyDestroyIcon(h : HICON);
procedure TimerStart;
procedure TimerEndAt(milliseconds : cardinal);

function GetCliptypeSymbol( Format: Cardinal ) : string;
{function IsPictureHint(hint : string) : boolean;
function GetPictureInt(hint : string) : integer;
}
function ShellExecute(F : THandle; command : string) : integer;

function GetAppPath : string;

function GetCF_HTML : UINT;
function GetCF_RICHTEXT : UINT;

procedure MySleep(milli : integer);


procedure SetDataFolderDefault;
procedure SetDataFolder(folder : string);

//procedure MySetTheme(h : Thandle);
function CloneIcon(inIcon : HICON) : HICON;
function UniSupported : boolean;


function ThreadAttach(TargetWindow: cardinal) : boolean;
procedure ThreadDetach();

const EXENAME_ERROR = '[error]';

{////////////////////}
{//}implementation{//}
{////////////////////}
uses Tlhelp32, UnitFrmMainPopup, Classes, Forms,
  UnitFrmClipboardManager, ShellAPI, UnitSpecialPaths, Dialogs,
  UnitPopupGenerate;

{
var SetWindowTheme: function(hwnd: HWND; pszSubAppName: LPCWSTR; pszSubIdList: LPCWSTR): HRESULT; stdcall;
  IsThemeActive: function: BOOL; stdcall;
}
var ThreadAttached : boolean;
    ThreadTarget : THandle;
    ThreadOurs : THandle;
function ThreadAttach(TargetWindow: cardinal) : boolean;
begin
    If (ThreadAttached) then begin
        showmessage('Error: Thread already attached');
        //self.Close;
    end;

    ThreadTarget := Windows.GetWindowThreadProcessId(TargetWindow, nil);
    ThreadOurs := Windows.GetCurrentThreadId();
    if (ThreadTarget <> ThreadOurs) then begin
        result := Windows.AttachThreadInput(ThreadTarget, ThreadOurs, true);
    end else begin
        result := true;
    end;
    ThreadAttached := result;
end;
procedure ThreadDetach();
begin
    if (ThreadTarget <> ThreadOurs) then begin
        Windows.AttachThreadInput(ThreadTarget, ThreadOurs, false);
    end;
    ThreadAttached := false;
end;


function UniSupported : boolean;
begin
    result := lo(Windows.GetVersion)>=5;
end;

function CloneIcon(inIcon : HICON) : HICON;
    var hIconNew : HICON;
        info : _ICONINFO;
    begin
        result := 0;

        // info from PCMagazine's TrayManager
        // Get information about the specified icon
        // Create a clone of the icon

        if (not Windows.GetIconInfo(inIcon, info)) then begin
            frmMainPopup.AppendLog('clone icon: ' + SysErrorMessage(GetLastError));
            EXIT;
        end;

        hIconNew := Windows.CreateIconIndirect(info);
        if (hIconNew = 0) then begin
            frmMainPopup.AppendLog('CreateIconIndirect failed: ' + SysErrorMessage(GetLastError));
        end;

        // Delete the info item's bitmaps -- otherwise we get a very
        // ugly memory leak
        if (not Windows.DeleteObject(info.hbmMask)) then begin
            frmMainPopup.AppendLog('DeleteObject failed: ' + SysErrorMessage(GetLastError));
        end;
        if (not Windows.DeleteObject(info.hbmColor)) then begin
            frmMainPopup.AppendLog('DeleteObject failed: ' + SysErrorMessage(GetLastError));
        end;
        result := hIconNew;

    end;
{
procedure MySetTheme(h : Thandle);
var dh : THandle;
    r : HRESULT;
begin
    if not ((Win32MajorVersion > 5) or
        ((Win32MajorVersion = 5) and (Win32MinorVersion = 1))) then EXIT;

    dh := LoadLibrary('UXTHEME.DLL');
    if dh <> 0 then begin
        SetWindowTheme := GetProcAddress(dh, 'SetWindowTheme');
    end;
    r := SetWindowTheme(h,'Toolbar','Toolbar');
    if r <> S_OK then begin
        //ShowMessage(SysErrorMessage(r));
    end;
end;
}
function RunCommandLine(command : string;priority : DWORD = NORMAL_PRIORITY_CLASS) : integer;
var  StartInfo  : _StartupInfoA;
     ProcInfo   : _PROCESS_INFORMATION;
     b : boolean;
begin
    // nil for the program name treats the program name
    // as a command line to execute
    FillChar(StartInfo, SizeOf(StartInfo), #0);
    FillChar(ProcInfo, SizeOf(ProcInfo), #0);
    StartInfo.cb := SizeOf(TStartupInfo);
    StartInfo.wShowWindow := SW_HIDE;

    b := CreateProcess(nil, PChar(command),
                nil, nil, False,
                CREATE_NEW_PROCESS_GROUP + priority,
                nil, nil, StartInfo, ProcInfo);
    result := 0;
    if (not b) then begin
        result := Windows.GetLastError();
    end;
end;

function ShellExecute(F : THandle; command : string) : integer;
var s : string;
    i : integer;
begin
    Windows.SetLastError(ERROR_SUCCESS);
    //
    // try to run it as a mail address if it fails
    // the first time - report any error otherwise

    // trim any whitespace - CRLFs
    s := trim(command);
    result := 0;
    if s = '' then EXIT;

    while s[length(s)] in [#13, #10] do begin
        s := Copy(s, 1, length(s) - 1);
    end;

    i := ShellAPI.ShellExecute(F,nil,
        PChar(s),nil,nil,SW_NORMAL);
    if (i <= 32) then begin

        if (pos('@',s) <> 0) then begin
            s := 'mailto:' + s;
            i := shellAPI.ShellExecute(F,nil,
                PChar(s),nil,nil,SW_NORMAL);
        end;
    end;

    result := 0;
    if (i <= 32) then begin
        result := Integer(RunCommandLine(s));
        //result := i;
        Windows.SetForegroundWindow(F);
        //ShowMessage(SysErrorMessage(i));
    end;
end;


procedure MySleep(milli : integer);
var i, j : integer;
begin
    i := Windows.GetTickCount;
    j := Windows.GetTickCount;
    while ((j - i) <milli) do begin
        Application.ProcessMessages;
        sleep(50);
        if (Application.Terminated) then EXIT;
        j := Windows.GetTickCount;
    end;
end;


var AppendCount : cardinal;
var AppendStr : string;
var TimerSnapshot : cardinal;
var AppPath : string;

function GetCF_HTML : UINT;
begin
    result := frmClipboardManager.GetCF_HTML;
end;
function GetCF_RICHTEXT : UINT;
begin
    result := frmClipboardManager.GetCF_RICHTEXT;
end;


function GetAppPath : string;
var par : string;
begin
    if AppPath = '' then begin
        par := '';
        if (ParamCount = 1) and (lowercase(ParamStr(1)) = '-data') then begin
                UnitMisc.SetDataFolderDefault;
        end else begin

            AppPath := IncludeTrailingPathDelimiter(
                ExtractFilePath(Application.ExeName)
            );
        end;
    end;
    result := AppPath;
end;

procedure SetDataFolderDefault;
begin
    AppPath := IncludeTrailingPathDelimiter(SpecialPaths.GetAppDataPath) + 'Arsclip\';
    ForceDirectories(AppPath);
end;

procedure SetDataFolder(folder : string);
begin
    AppPath := IncludeTrailingPathDelimiter(folder);
end;

{
function GetPictureInt(hint : string) : integer;
begin
    result := StrToInt( midstr(hint,7,length(hint)-7) );
end;
function IsPictureHint(hint : string) : boolean;
begin
    result := uppercase(leftstr(hint,6)) =  '[PICD=';
end;
}

procedure TimerStart;
begin
    TimerSnapshot := Windows.GetTickCount;
end;

procedure TimerEndAt(milliseconds : cardinal);
begin
    while (Windows.GetTickCount - TimerSnapshot) < milliseconds do begin
        // this procedures purposely stalls the program, so ProcessMessages
        // is NOT used here
    end;
end;

//
// I want "free" object references to be nil for easy testing
// Notes: The .Free documenation is very misleading, this function
// make it 100% clear
//
// - Free only if not null
// - Make null when freed
//

procedure MyFree(var o);
begin
    if (Pointer(o) <> nil) then begin
        TObject(o).free;
        Pointer(o) := nil;
    end;
end;


procedure MyDestroyIcon(h : HICON);
begin
    if (h <> 0) then begin
        UnitMisc.AppendLog('^DestroyIcon^');
        if not Windows.DestroyIcon(h) then begin
            UnitMisc.AppendLog('Icon not destroyed', true);
        end;
    end;
end;


//
// Only returns Empty String when the handle is not found,
// not on an error
//
function WindowHandleToEXEName(handle : THandle) : string;
var snap : THandle;
    pe : tagPROCESSENTRY32;
    pid : THandle;
    found : boolean;
begin
    Windows.SetLastError(ERROR_SUCCESS);

    result := '';
    if (handle = 0) then begin
        EXIT;
    end;
    snap := TLHelp32.CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0);
    if (snap = Cardinal(-1)) then begin
        result := EXENAME_ERROR;
        EXIT;
    end;

    Windows.GetWindowThreadProcessId(handle, pid);
    pe.dwSize := Sizeof(pe);
    found := TLHelp32.Process32First(snap, pe);

    while found do begin
        if (pe.th32ProcessID = pid) then begin
            result := String(pe.szExeFile);
            break;
        end;

        found := TLHelp32.Process32Next(snap, pe);
    end;
    CloseHandle(snap);
end;


function DupHandle(h : Thandle; var sizeh : cardinal; SizeLimit : boolean = false) : Thandle;
var sz : Cardinal;
    newh : Thandle;
    p1, p2 : pointer;
    b : boolean;
begin
    Windows.SetLastError(ERROR_SUCCESS);

    result := 0;
    if (h = 0) then begin
        AppendLog('<DupHandle - Empty handle>');
        EXIT;
    end;
    //
    // make sure size is non-zero
    // and not bigger than size restrictions
    //
    sz := Windows.GlobalSize(h);
    if (sz = 0) then begin
        AppendLog('<DupHandle - GlobalSize failed >', true);
        sizeh := 0;
        EXIT;
    end;

    // abort if too big
    // make sure to return size of item
    if (SizeLimit) and (sz > sizeh) then begin
        AppendLog('<DupHandle - too big, size limit>');
        sizeh := sz;
        EXIT;
    end else begin
        sizeh := sz;
    end;

    //
    // Lock and copy
    //

    newh := Windows.GlobalAlloc(GMEM_MOVEABLE or GMEM_ZEROINIT, sizeh);
    if (newh=0) then begin
        AppendLog('<DupHandle - GlobalAlloc failed >', true);
        EXIT;
    end;
    p1 := Windows.GlobalLock(newh);
    if (p1=nil) then begin
        AppendLog('<DupHandle lock1 >', true);
        EXIT;
    end;

    p2 := Windows.GlobalLock(h);
    if (p2=nil) then begin
        Windows.GlobalUnlock(newh);
        AppendLog('<DupHandle lock2 >', true);
        EXIT;
    end;
    
    MoveMemory(p1, p2, sizeh);

    //
    // This should never occur, maybe hardware failure might cause this
    //
    b := CompareMem(p1, p2, sizeh);
    AppendLog('<DupHandle compare=>' + BoolToStr(b));
    if ( not b) then begin
        Application.ShowException(Exception.Create('DupHandle - Failed to copy memory'));
        Application.Terminate;
    end;


    Windows.GlobalUnlock(h);
    Windows.GlobalUnlock(newh);

    result := newh;
end;

procedure AppendLog( s : string; IncludeLastError : boolean = false);
begin
    if (IncludeLastError) then begin
        s := s + ' : ' + SysErrorMessage(GetLastError);
    end;
    if (AppendStr = '') then begin
        AppendStr := TimeToStr(Now) + ': ' + s;
    end else begin
        AppendStr :=  TimeToStr(Now) + ': ' + s +  #13#10 + AppendStr;
    end;


    AppendCount := (AppendCount + 1) mod 5;
    if (AppendCount = 0) then begin
        FrmMainPopup.AppendLog(AppendStr);
        AppendStr := '';
    end;

    //showmessage(s);
end;
procedure FlushLog;
begin
    FrmMainPopup.AppendLog(AppendStr);
    AppendCount := 0;
end;

procedure ApplicationException(Sender: TObject; E: Exception);
begin
    ShowMessage(
        e.Message + #13#10 +
        '[Debug.txt] log created.'
    );

    AppendLog(e.Message);

    DumpLog('Debug.txt');
end;


procedure DumpLog(filename : string);
begin
    FlushLog;
    filename := UnitMisc.GetAppPath + filename;

    FrmMainPopup.Memo1.Lines.SaveToFile(filename);
end;


function GetCliptypeSymbol(Format : Cardinal) : string;
begin
    case Format of
        CF_UNICODETEXT : begin
            result := '[U]';
        end;
        CF_HDROP : begin
            result := '[F]' ;
        end;
        CF_TEXT : begin
            result := '[T]' ;
        end;
        CF_DIB : begin
            result := '[P]' ;
        end;
    end;
    if (Format = UnitMisc.GetCF_RICHTEXT) then begin
        result := '[R]' ;
    end else if (Format = UnitMisc.GetCF_HTML) then begin
        result := '[H]' ;
    end;
end;

initialization
begin
    AppendCount := 1;
end;
end.
