unit UnitPopupGenerate;
{
    Purpose:
        Move the popup generation logic into a separate file

    Notes:
        Used a dirty trick by just replacing FrmMainPopup with a subclass

}

{///////////////}
{//}interface{//}
{///////////////}
uses Menus, UnitFrmMainPopup, UnitMenuItemTagdata, UnitClipQueue, Windows,
  Graphics, UnitMisc, Classes;

type TPopupGenerator = class(TFrmMainPopup)
        procedure FormCreate(Sender: TObject); overload;

    private
        s : string;
        ci : TClipItem;
        lastaddcount : integer;
        popup : TPopupMenu;

        lastHI : HICON;
        permshown : boolean;
        allitemsshown : boolean;
        
        iconcache : TStringList;

        NextItemHasBreak : boolean;
        
        procedure AddToPopup(mi : TMenuItem);
        procedure ReserveShortcut();
        procedure AddProgramOptions;
        function GetNewMenuItem : TMenuItem;
        function GetNewTagData(m : TMenuItem) : TMenuItemTagData;
        function GetNewLine : TMenuITem;
        procedure AddTextItems;
        procedure AddFullConfigured;
        procedure AddSystem;
        procedure AddSwitchToSubmenu;
        procedure AddSwitchToItems;
        procedure AddMisc;
        procedure AddCurrentClipboard;
        procedure AddLast;
        procedure AddOtherItems;
        procedure AddRecentlyRemoved;
        procedure AddCancel;
        procedure AddCurrentPermanentItems;
        procedure AddAllItemsSubmenu;
        procedure AddSelected;
        procedure AddSearch;
        procedure AddBreak;
        procedure DisableClickEvent(sender : TObject);
        procedure NoRightClickEvent(sender : TObject);

        procedure AddRecentlyClicked;
    protected
        procedure PopulatePopupANSI(p : TPopupMenu); override;
    public
        procedure FakePopupHotkey;

end;

var
  FrmMainPopup: TPopupGenerator;

{////////////////////}
{//}implementation{//}
{////////////////////}

uses UnitFrmConfig, UnitPaste, UnitFrmSysTrayMenu,
  UnitOtherQueue, SysUtils, UnitFrmPermanentNew, UnitFrmClipboardManager, Forms, Messages;


function TPopupGenerator.GetNewLine: TMenuITem;
var td : TMenuItemTagData;
begin
    result := newline();
    td := GetNewTagData(result);
    td.itemtype := IT_LINE;
end;

function TPopupGenerator.GetNewMenuItem : TMenuItem;
begin
    result := TmenuItem.create(popup);
    result.tag := 0;
    result.ImageIndex := 0;
    result.Hint := '';
    result.Checked := false;
    self.GetNewTagData(result); 
    //result.Bitmap := nil;
end;

function TPopupGenerator.GetNewTagData(m : TMenuItem) : TMenuItemTagData;
begin
    if m.tag = 0 then
        m.Tag := TagDataList.NewTagData;
    result := TagDataList.Get(m.tag);
end;



procedure TPopupGenerator.AddToPopup(mi : TMenuItem);
var s : string;
    mitd : TMenuItemTagData;
    i : integer;
begin
    if mi.Tag = 0 then begin
        mi.Tag := TagDataList.NewTagData;
    end;
    mitd := TagDataList.Get(mi.tag);

    // generate an accelerator key for all main popup
    // items not already marked with '&'

    if (mi.Count = 0) and (not mi.IsLine)  and
        (FrmConfig.cbUseKeyboard.Checked) and
        (pos('&',mi.caption) = 0) and
        (not (mitd.itemtype = IT_MENU))
    then begin
        // reserve spots for the text/non-text menu items
        // - these can have less items than their full size
        i := ClipQueue.GetQueueSize + OtherQueue.GetQueueSize;


        case (mitd.itemtype) of
        IT_MENU: i := 0;
        IT_OTHER: i := ClipQueue.GetQueueSize;
        end;


        s := self.GetAccelerator( lastaddcount + i);
        mitd.prefix := '&'+ s + ': ';
        inc(lastaddcount);
    end;
    if NextItemHasBreak  then begin
        NextItemHasBreak := false;
        mi.Break := mbBarBreak;
    end;

    popup.items.add(mi);
end;





procedure TPopupGenerator.DisableClickEvent(sender: TObject);
begin
    FrmSysTrayMenu.ToggletDisabledEXE(targetEXE);
end;

procedure TPopupGenerator.NoRightClickEvent(sender: TObject);
begin
    frmSysTrayMenu.ToggleNoRightClickEXE(targetEXE);
end;


procedure TPopupGenerator.FakePopupHotkey;
var msg : TWMHotKey;
begin
    msg.HotKey  := FrmConfig.GetHotKeyID;
    self.WMHotKey(msg);
end;

procedure TPopupGenerator.FormCreate(Sender: TObject);
begin
    inherited;
    iconcache := TStringList.Create;
    iconcache.Duplicates := dupAccept;

end;

procedure TPopupGenerator.AddProgramOptions;
var top, menu, submenu  : TMenuItem;
  mitd: TMenuItemTagData;
begin

    top := self.GetNewMenuItem;
    top.caption := 'Program options: [' + targetEXE +']';
    mitd := GetNewTagData(top);

    if lastHI <> 0 then begin
        DestroyIcon(lastHI);
    end;
    mitd.icon := FrmClipboardManager.GetClipboardOwnerIcon(self.GetNextWindow);
    lastHI := mitd.icon;
    mitd.itemtype := IT_PROGRAM;

    //top.Add(NewLine());

   {menu := self.GetNewMenuItem;
    menu.caption := '&Paste Method ' +  targetEXE;
    menu.Enabled := false;
    top.add(menu);
    }
    submenu := self.GetNewMenuItem;
    submenu.Caption := 'CTRL+V';
    submenu.OnClick := FrmSysTrayMenu.MethodMenuItemClickEvent;
    mitd := GetNewTagData(submenu);
    mitd.itemtype := IT_SYS_METHOD;
    mitd.itemindex := Integer(PASTE_CTRL_V);
    top.Add(submenu);

    submenu := self.GetNewMenuItem;
    submenu.Caption := 'SHIFT+INSERT';
    submenu.OnClick := FrmSysTrayMenu.MethodMenuItemClickEvent;
    mitd := GetNewTagData(submenu);
    mitd.itemtype := IT_SYS_METHOD;
    mitd.itemindex := Integer(PASTE_SHIFT_INS);

    top.Add(submenu);

    submenu := self.GetNewMenuItem;
    submenu.Caption := 'Mimic Typing';
    submenu.OnClick := FrmSysTrayMenu.MethodMenuItemClickEvent;
    mitd := GetNewTagData(submenu);
    mitd.itemtype := IT_SYS_METHOD;
    mitd.itemindex := Integer(PASTE_MIMIC);
    top.Add(submenu);



    submenu := self.GetNewMenuItem;
    submenu.Caption := 'Clipboard Only';
    submenu.OnClick := FrmSysTrayMenu.MethodMenuItemClickEvent;
    mitd := GetNewTagData(submenu);
    mitd.itemtype := IT_SYS_METHOD;
    mitd.itemindex := Integer(PASTE_CLIPBOARD);
    top.Add(submenu);

    submenu := self.GetNewMenuItem;
    submenu.Caption := 'Default';
    submenu.OnClick := FrmSysTrayMenu.MethodMenuItemClickEvent;
    mitd := GetNewTagData(submenu);
    mitd.itemtype := IT_SYS_METHOD;
    mitd.itemindex := Integer(PASTE_DEFAULT);
    top.Add(submenu);


    top.Items[integer(Paste.GetPasteMethod(targetEXE))].Checked := true;

    top.add(GetNewLine());

    menu := self.GetNewMenuItem;
    menu.caption := 'Disable Monitoring';
    menu.Hint := targetEXE;
    menu.OnClick := self.DisableClickEvent;
    menu.Checked := FrmSysTrayMenu.IsDisabledEXE(targetEXE);
    menu.ImageIndex := 0;
    top.add(menu);


    menu := self.GetNewMenuItem;
    menu.caption := 'Disable Right-Click Trigger';
    menu.Hint := targetEXE;
    menu.OnClick := self.NoRightClickEvent; 
    menu.Checked := FrmSysTrayMenu.IsNoRightClickEXE(targetEXE);
    menu.ImageIndex := 0;
    top.add(menu);


    //top.Add(menu);

    AddToPopup(top);
end;






procedure TPopupGenerator.PopulatePopupANSI(p : TPopupMenu);
var i, cnt : integer;
    ic : TIcon;
    linebefore : boolean;
    mot : TMenuOrderType;
    m : TMenuItem;
begin

    for i := 0 to iconcache.count - 1 do begin
        ic := TIcon(iconcache.Objects[i]);
        myfree(ic);
    end;
    iconcache.Clear;

    self.popup := p;
    popup.Items.Clear;
    TagDataList.flush;
    self.lastaddcount := 0;


    if self.UseFormMode then begin
        m := GetNewMenuItem;
        m.caption := '&- >> Next Field';
        m.AutoCheck  := true;
        m.OnClick := self.NextFieldModeMenuItemClickEvent;
        m.Bitmap := self.imgEdit.Picture.Bitmap;
        AddToPopup(m);
        AddToPopup(Self.GetNewLine);
    end;

    // find out if permanent items are shown
    self.permshown := false;
    self.allitemsshown := false;
    for I := 0 to FrmConfig.lvOrder.items.Count - 1 do begin
        mot := TMenuOrderType(frmConfig.lvOrder.Items[i].data);
        if frmConfig.lvOrder.Items[i].Checked then
            case mot of
                MOT_PERMANENT : permshown := true;
                MOT_ALLITEMS :  self.allitemsshown := true;
            end;
    end;

    for I := 0 to FrmConfig.lvOrder.items.Count - 1 do begin
        linebefore := (FrmConfig.lvOrder.items[i].SubItems[0] = 'Yes');

        if frmConfig.lvOrder.Items[i].Checked or self.FullMode then begin
            mot := TMenuOrderType(frmConfig.lvOrder.Items[i].data);
            {case mot of
                MOT_MENU: if ClipQueue.GetQueueCount= 0 then linebefore := false;
                MOT_OTHER: if OtherQueue.GetQueueCount= 0 then linebefore := false;
                MOT_PERMANENT: if FrmPermanent.GetCount= 0 then linebefore := false;
            end;}
            if linebefore then begin
                AddToPopup(Self.GetNewLine);
            end;
            cnt := popup.Items.Count;
            case mot of
                MOT_MENU : self.AddTextItems;
                MOT_OTHER : if frmConfig.cbSeparateNontextList.checked then self.AddOtherItems;
                MOT_CANCEL : self.AddCancel;
                MOT_LAST : self.AddLast;
                MOT_CURRENT : self.AddCurrentClipboard;
                MOT_SELECTED : self.AddSelected;
                MOT_REMOVED : self.AddRecentlyRemoved;
                MOT_PROGRAM : self.AddProgramOptions;
                MOT_SYSTEM : self.AddSystem;
                MOT_MISC : self.AddMisc;
                MOT_PERMANENT : begin
                    // don't show in full mode, unless configured to show normally
                    if not (self.FullMode and (not frmConfig.lvOrder.Items[i].Checked)) then begin
                        self.AddCurrentPermanentItems;
                    end;
                end;
                MOT_SWITCH :
                    if frmconfig.cbPermanentItemsSubmenu.checked then
                        self.AddSwitchToSubmenu
                    else
                        self.AddSwitchToItems;
                MOT_FULL : self.AddFullConfigured;
                MOT_ALLITEMS : self.AddAllItemsSubmenu;
                MOT_CLICKED: self.AddRecentlyClicked;
                MOT_SEARCH: self.AddSearch;
                MOT_BREAK: if MOT = MOT_BREAK then self.AddBreak; // don't add breaks in Full Mode
            end;
            // remove lines for items that generate no new menuitems
            if linebefore and (cnt = popup.Items.count) then begin
                popup.Items.Delete(popup.items.count-1);
            end;
        end;
    end;


    {if (self.PopupUseKB) then begin
        popup.AutoHotkeys := maAutomatic;
    end else begin
        popup.AutoHotkeys := maManual;
    end;
    Popup.OwnerDraw := false;
    }

    //
    // Recreate the menuitems
    //
end;

procedure TPopupGenerator.ReserveShortcut;
begin
    inc(lastaddcount);
end;

procedure TPopupGenerator.AddTextItems;
var i : integer;
    m : TMenuItem;
  mitd: TMenuItemTagData;

begin
    for i := 0 to (ClipQueue.GetQueueCount - 1) do begin
        m := GetNewMenuItem;
        s := ClipQueue.GetItemText(i);
        m.Caption := GetTrimmedItem(s);
        ci := ClipQueue.GetClipItem(i);

        m.OnClick := self.MenuItemClickEvent;


        mitd := GetNewTagData(m);
        mitd.name := s;
        mitd.caption := m.caption;
        mitd.clip := ci;
        mitd.itemtype := IT_MENU;
        mitd.itemindex := i;
        if (FrmConfig.cbUseKeyboard.Checked) then begin
            mitd.prefix := '&' + GetAccelerator(i) + ': ';
        end;

        AddToPopup(M);
    end;
end;


procedure TPopupGenerator.AddOtherItems;
var i : integer;
    m : TMenuItem;
  mitd: TMenuItemTagData;
begin

    for i := (OtherQueue.GetQueueCount - 1) downto 0 do begin
        m := GetNewMenuItem;
        mitd := GetNewTagData(m);

        ci := OtherQueue.GetClipItem(i);
        m.Caption := ci.GetAsText  + ': size ' +
            IntToStr(Trunc(ci.GetDataSize  / 1024)) +
            ' KB';

        m.OnClick := OtherItemClickEvent;



        mitd.name := m.caption;
        mitd.caption := m.caption;
        mitd.itemtype := IT_OTHER;
        mitd.itemindex := i;
        mitd.clip := ci;

        AddToPopup(m);
    end;
end;

procedure TPopupGenerator.AddBreak;
var m : TMenuItem;
begin
    // Since this is not a visual menu item,
    // the AddToPopup function is cirumvented


    self.NextItemHasBreak := true;
   // popup.Items[popup.Items.Count-1].Break := mbBarBreak;
end;

procedure TPopupGenerator.AddCancel;
var
  m: TMenuItem;
begin
    m := GetNewMenuItem;
    m.caption := 'Cancel';
    m.OnClick := self.CancelMenuItemClickEvent;
    AddToPopup(m);
end;
procedure TPopupGenerator.AddLast;
var m : TMenuItem;
    mitd : TMenuItemTagData;
begin
    if LastStringSelected = '' then begin
        self.ReserveShortcut;
        EXIT;
        // even in full mode, don't show it empty
        // but do reserve the shortcut letter so keys don't "jump"
        // when Last is empty
    end;
    
    m := GetNewMenuItem;
    mitd := GetNewTagData(m);
    mitd.itemtype := IT_LAST;

    m.caption := 'Last: ' + GetTrimmedItem(LastStringSelected);
    m.OnClick := self.LastMenuItemClickEvent;

    mitd := GetNewTagData(m);
    mitd.prefix := 'Last: ';
    mitd.text := LastStringSelected;
    AddToPopup(m);
end;
procedure TPopupGenerator.AddCurrentClipboard;
var m : TMenuItem;
  mitd: TMenuItemTagData;
begin
    m := GetNewMenuItem;
    mitd := GetNewTagData(m);
    mitd.itemtype := IT_CURRENT;
    self.CurrentClipboard.GetClipboardItem(0);
    if (self.CurrentClipboard.HasText) then begin
        s := self.CurrentClipboard.GetAsText;
        m.caption := GetTrimmedItem('Current: ' + self.CurrentClipboard.GetAsText);
    end else begin
        s := self.CurrentClipboard.GetFormatName;
        if (s = '') then begin
            s :=  GetTrimmedItem(self.CurrentClipboard.GetAsText);
        end;
        m.caption := GetTrimmedITem('Current: ' + s);
    end;
    m.OnClick := self.CurrentMenuItemClickEvent;
    AddToPopup(m);
end;

procedure TPopupGenerator.AddSearch;
var m, subm : TMenuItem;
    mitd: TMenuItemTagData;
begin
    m := GetNewMenuItem;
    m.Caption := 'Search';
    m.OnClick := self.SearchMenuItemClickEvent;
    m.Bitmap := imgEdit.Picture.Bitmap;

    mitd := GetNewTagData(m);
    mitd.itemtype := IT_SEARCH;
    mitd.text := 'Search';
    AddToPopup(m);
end;

procedure TPopupGenerator.AddSelected;
var m : TMenuItem;
    mitd: TMenuItemTagData;
begin
    if self.TargetSelText = '' then begin
        //self.ReserveShortcut;
        EXIT;
    end;
    
    m := GetNewMenuItem;
    mitd := GetNewTagData(m);
    mitd.itemtype := IT_SELECTED;
    m.Caption := GetTrimmedItem('Edit Selected: ' + self.TargetSelText);
    m.OnClick := SelectedAllMenuItemClickEvent;

    //mitd.prefix := 'Edit Selected: ';
    mitd.prefix := '';    
    mitd.text := 'Edit Selected: ' + self.TargetSelText; 
    AddToPopup(m);
end;


procedure TPopupGenerator.AddRecentlyRemoved;
var i : cardinal;
    m, subm : TMenuItem;
  mitd: TMenuItemTagData;
begin
    if TempClipQueue.GetQueueCount = 0 then begin
        //self.ReserveShortcut;
    end else begin
        subm := GetNewMenuItem;
        subm.Caption := 'Recently Removed Clips';
        subm.OnClick := nil;
        subm.Bitmap := imgRemoved.Picture.Bitmap;

        mitd := GetNewTagData(subm);
        mitd.itemtype := IT_POPUP;
         for i := 0 to (TempClipQueue.GetQueueCount - 1) do begin
            m := GetNewMenuItem;
            s := TempClipQueue.GetItemText(i);
            ci := TempClipQueue.GetClipItem(i);
            m.Caption  :=  GetTrimmedItem(s);

            mitd := GetNewTagData(m);

            mitd.name := s;
            mitd.itemindex := i;
            mitd.itemtype := IT_TEMP;

            mitd.caption := m.caption;
            mitd.clip := ci;

            m.OnClick := self.TempMenuItemClickEvent;
            subm.Add(M);

            if i = (TempClipQueue.GetQueueCount - 1) then begin
                subm.Add(GetNewLine());
            end;
        end;
        if subm.Count = 0 then begin
            m := GetNewMenuItem;
            m.Enabled := false;
            m.Caption := 'none';

            subm.Add(m);
            subm.Add(GetNewLine);
        end;
        m := GetNewMenuItem;
        m.Caption := 'Removed Clips ...';
        m.Bitmap := FrmMainPopup.imgRemoved.Picture.Bitmap;
        m.OnClick := frmSysTrayMenu.RemovedItemsMenuItemClickEvent;
        m.ImageIndex := 0;
        subm.Add(m);
        AddToPopup(subm);
    end;
end;

procedure TPopupGenerator.AddRecentlyClicked;
var i : cardinal;
    m, subm : TMenuItem;
  mitd: TMenuItemTagData;
begin
    if ClickedclipQueue.GetQueueCount = 0 then begin
        //self.ReserveShortcut;
    end else begin
        subm := GetNewMenuItem;
        subm.Caption := 'Recently Clicked';
        subm.OnClick := nil;
        subm.Bitmap := imgFlush.Picture.Bitmap;
        mitd := GetNewTagData(subm);
        mitd.itemtype := IT_POPUP;


         for i := 0 to (ClickedclipQueue.GetQueueCount - 1) do begin
            m := GetNewMenuItem;
            s := ClickedclipQueue.GetItemText(i);
            ci := ClickedclipQueue.GetClipItem(i);
            m.Caption  :=  GetTrimmedItem(s);

            mitd := GetNewTagData(m);

            mitd.name := s;
            mitd.itemindex := i;
            mitd.itemtype := IT_CLICKED;

            mitd.caption := m.caption;
            mitd.clip := ci;

            m.OnClick := self.TempMenuItemClickEvent;
            subm.Add(M);

            if i = (ClickedclipQueue.GetQueueCount - 1) then begin
                subm.Add(GetNewLine());
            end;
        end;
        if subm.Count = 0 then begin
            m := GetNewMenuItem;
            m.Enabled := false;
            m.Caption := 'none';

            subm.Add(m);
            subm.Add(GetNewLine);
        end;
        {m := GetNewMenuItem;
        m.Caption := 'Removed Items';
        m.Bitmap := FrmMainPopup.imgRemoved.Picture.Bitmap;
        m.OnClick := frmSysTrayMenu.RemovedItemsMenuItemClickEvent;
        m.ImageIndex := 0;
        subm.Add(m);}
        AddToPopup(subm);
    end;
end;

procedure TPopupGenerator.AddMisc;
var m, subm : TMenuItem;
    mitd : TMenuItemTagData;
begin
    subm := GetNewMenuItem;
    mitd := GetNewTagData(subm);
    mitd.itemtype := IT_POPUP;

    m := GetNewMenuItem;
    m.Caption := 'Edit History...';
    m.OnClick := self.EditHistoryItemClickEvent;
    m.Bitmap := imgEdit.Picture.Bitmap;
    subm.Add(m);
    subm.Add(GetNewLine);

    m := GetNewMenuItem;
    m.caption := 'Paste All Text';
    m.OnClick := self.PasteAllMenuItemClickEvent;
    m.Bitmap := self.imgPasteAll.Picture.Bitmap;
    subm.Add(m);

    m := GetNewMenuItem;
    m.caption := 'Paste Selected Text...';
    m.OnClick := self.PasteSelectedMenuItemClickEvent;
    m.Bitmap := self.imgPaste.Picture.Bitmap;
    subm.Add(m);

    subm.Add(GetNewLine);

    m := GetNewMenuItem;
    m.Checked := self.UseFormMode;
    m.Caption := 'Form Mode';
    m.AutoCheck  := true;
    m.Checked := self.UseFormMode;
    m.OnClick := self.FormModeMenuItemClickEvent;
    m.Bitmap := self.imgEdit.Picture.Bitmap;
    subm.Add(m);

    m := GetNewMenuItem;
    m.caption := '>> Next Field';
    m.AutoCheck  := true;
    m.OnClick := self.NextFieldModeMenuItemClickEvent;
    m.Bitmap := self.imgEdit.Picture.Bitmap;
    subm.Add(m);

    subm.bitmap := Self.ImgMisc.Picture.Bitmap;
    subm.Caption := 'Misc';
    AddToPopup(subm);
end;

procedure TPopupGenerator.AddCurrentPermanentItems;
var i : integer;
    m : TMenuItem;
  mitd: TMenuItemTagData;
begin
     for i := 0 to frmPermanent.GetCount - 1 do begin
        m := GetNewMenuItem;
        m.Caption := FrmPermanent.GetItemName(i);
        m.OnClick := self.PermanentMenuItemClickEvent;

        m.Bitmap := FrmMainPopup.ImgEdit.Picture.Bitmap;

        mitd := GetNewTagData(m);
        mitd.PermanentGroupID := -1;
        mitd.itemindex := i;
        mitd.text := FrmPermanent.GetItemText(i);
        mitd.caption := m.Caption;
        mitd.itemtype := IT_PERMANENT;
        
        AddToPopup(m);
        {
        mitd.itemtype := IT_PERMANENT;
        if FrmPermanent.IsComplexItem(mitd.text) then begin
            mitd.clip := FrmPermanent.GetComplexItem(mitd.text);
        end;
        }
    end;
end;



procedure TPopupGenerator.AddSwitchToItems;
var i : integer;
    m : TMenuItem;
  mitd: TMenuItemTagData;
begin
    s := lowercase(frmPermanent.GetPermanentPath);
    for i := 0 to frmPermanent.PermFoldersGetCount - 1 do begin
        if (lowercase(frmPermanent.PermFoldersGetItem(i)) <> s) then begin
            m := GetNewMenuItem;
            m.Caption := 'Switch to ' + frmPermanent.PermFoldersGetItem(i);
            m.OnClick := self.PermanentGroupItemClickEvent;
            m.Bitmap := imgRightArrow.Picture.Bitmap;

            mitd := GetNewTagData(m);
            mitd.PermanentGroupID := i;
            mitd.itemtype := IT_SWITCH;

            AddToPopup(m);
        end;
    end;
end;
procedure TPopupGenerator.AddSwitchToSubmenu;
var i : integer;
    m, subm : TMenuItem;
  mitd: TMenuItemTagData;
begin
    subm := GetNewMenuItem;
    subm.Caption := 'Permanent Groups ...';
    subm.Bitmap := imgPermCasc.Picture.Bitmap;
    mitd := GetNewTagData(subm);
    mitd.itemtype := IT_POPUP;

    s := lowercase(frmPermanent.GetPermanentPath);

    for i := 0 to frmPermanent.PermFoldersGetCount - 1 do begin
        m := GetNewMenuItem;
        mitd := GetNewTagData(m);
        m.Caption := FrmPermanent.PermFoldersGetItem(i);
        mitd.caption := m.caption;
        mitd.PermanentGroupID := i;
        mitd.itemtype := IT_SWITCH;

        if (lowercase(m.caption) <> s) then begin
            m.Caption := 'Switch to ' + m.caption;
        end else begin
            m.Caption := 'Switch to ' + m.caption + ' (Current)';
        end;
        mitd.caption := m.caption;
        m.OnClick := self.PermanentGroupItemClickEvent;
        subm.Add(m);

    end;


    m := GetNewLine;
    subm.Add(m);
    m := GetNewMenuItem;
    m.Caption := 'Edit Permanent Clips';
    m.OnClick := frmSysTrayMenu.SwitchMenuItemClickEvent;
    m.Bitmap := imgPermCasc.Picture.Bitmap;
    subm.Add(m);

    AddToPopup(subm);
end;

procedure TPopupGenerator.AddAllItemsSubmenu;
var i, j : integer;
    m, main, subm2 : TMenuItem;
    mitd : TMenuItemTagData;
    count : integer;
begin
    main := nil;
    count := 0;
    if (not FrmConfig.cbShowAllAsSubmenus.checked) then begin
        main := GetNewMenuItem;
        main.Hint := '';
        main.Caption := 'All Items...';
        main.Bitmap := imgPermCasc.Picture.Bitmap;
        main.OnClick := nil;
        main.ImageIndex := 0;
        mitd := GetNewTagData(main);
        mitd.itemtype := IT_POPUP;
    end;
    s := lowercase(frmPermanent.GetPermanentPath);

    FrmPermanent.PermFolderPush;
    for i := 0 to frmPermanent.PermFoldersGetCount - 1 do begin
        if (lowercase(frmPermanent.PermFoldersGetItem(i)) <> s) or (not self.permshown) then begin
            m := GetNewMenuItem;
            m.Caption := frmPermanent.PermFoldersGetItem(i);
            m.ImageIndex := 0;
            mitd := GetNewTagData(m);
            mitd.itemtype := IT_POPUP;

            FrmPermanent.SetPermanentPath(m.Caption);
            count := FrmPermanent.GetCount;
            for j := 0 to FrmPermanent.GetCount - 1 do begin
                subm2 := GetNewMenuItem;
                subm2.Caption := FrmPermanent.GetItemName(j);
                subm2.OnClick := self.PermanentMenuItemClickEvent;
                subm2.ImageIndex := 0;

                mitd := GetNewTagData(subm2);
                mitd.itemtype := IT_PERMANENT;
                mitd.PermanentGroupID := i;
                mitd.itemindex := j;
                mitd.text := FrmPermanent.GetItemText(j);
                mitd.name := subm2.caption;
                subm2.Bitmap := FrmMainPopup.ImgEdit.Picture.Bitmap;

                m.Add(subm2);
            end;

            m.OnClick := nil;
            if (count >0) then begin
                if (FrmConfig.cbShowAllAsSubmenus.checked) then begin
                    AddToPopup(m);
                end else begin
                    main.Add(m);
                end;
            end;
        end;

    end;

    FrmPermanent.PermFolderPop;
    if (not FrmConfig.cbShowAllAsSubmenus.checked) then begin
        if (main.Count>0) then begin
            m := GetNewLine;
            main.Add(m);
            m := GetNewMenuItem;
            m.Caption := 'Edit Permanent Items';
            m.OnClick := frmSysTrayMenu.SwitchMenuItemClickEvent;
            m.Bitmap := FrmMainPopup.imgPermCasc.Picture.Bitmap;
            main.Add(m);

            AddToPopup(main);
        end;
    end;
end;

procedure TPopupGenerator.AddSystem;
var
    m : TMenuItem;

    procedure FixMenuItems(m : TMenuItem);
    var i : integer;
        mitd : TMenuItemTagData;
        ic : TIcon;
    begin
        for i  := 0 to m.Count -1 do begin
            //mitd.icon := m.items[i].Bitmap.
            mitd := GetNewTagData(m.items[i]);

            if m.items[i].ImageIndex >= 0 then begin
                ic := TIcon.Create;
                //frmSysTrayMenu.ilPopupIcons.GetBitmap(m.items[i].ImageIndex,m.items[i].Bitmap);
                frmSysTrayMenu.ilPopupIcons.GetIcon(m.items[i].ImageIndex, ic);

                mitd.icon := ic.Handle;
                iconcache.AddObject('',ic);
            end;
            if m.items[i].IsLine then begin
                mitd.itemtype := IT_LINE;
            end else begin
                mitd.itemtype := IT_SYSTEM;
            end;

            if m.Items[i].Count <> 0 then begin
                mitd.itemtype := IT_POPUP;
                FixMenuItems(m.items[i]);
            end;
        end;
    end;
var mitd : TMenuItemTagData;
begin
    m := GetNewMenuItem;


    mitd := GetNewTagData(m);

    //mitd.icon := application.Icon.Handle;
    //mitd.icon :=FrmMainPopup.imgIcon.Picture.icon.handle;



    frmSysTrayMenu.ShowPopup(m);

    FixMenuItems(m);

    m.Caption := 'System';
    m.Hint := '';
    m.bitmap := self.imgAICon.Picture.Bitmap;
    mitd.itemtype := IT_POPUP;
    m.OnClick := nil;
    m.ImageIndex := 0;

   // mitd.icon := FrmMainPopup.imgIcon.Picture.icon.handle;
    AddToPopup(m);

end;

procedure TPopupGenerator.AddFullConfigured;
var
    m : TMenuItem;
    mitd : TMenuItemTagData;
begin
    m := GetNewMenuItem;
    if FrmConfig.cbUseKeyboard.Checked then begin
        m.caption := '&` ';
    end;
    if (self.FullMode = true) then begin
        m.Caption := m.caption + 'Configured Mode';
        m.Bitmap := self.imgConfMode.Picture.Bitmap;
    end else begin
        m.Caption := m.caption + 'Full Mode';
        m.Bitmap := self.imgFull.Picture.Bitmap;
    end;
    m.OnClick := self.FullModeMenuItemClickEvent;

    mitd := self.GetNewTagData(m);
    mitd.itemtype := IT_FULL;

    AddToPopup(m);

end;



end.
